class Memory:

    def __init__(self, scheduler: Scheduler, gamepak: GamePak, bios: bytes | bytearray | Iterable[int]):
        self.scheduler = scheduler
        self.cpu: CPU | None = None
        self.io: IO | None = None
        self.bios = array('B', bios.ljust(MemoryRegion.BIOS_SIZE))
        self.ewram = array('B', [0] * MemoryRegion.EWRAM_SIZE)
        self.iwram = array('B', [0] * MemoryRegion.IWRAM_SIZE)
        self.gamepak = gamepak
        self.bios_last_opcode = 0
        self.wait_control = WaitstateControlRegister()
        self._init_access_times()

    def read_32(self, address: int, access_type: MemoryAccess) -> int:
        return self._read_32_internal(address, access_type)

    def read_32_ror(self, address: int, access_type: MemoryAccess) -> int:
        value = self.read_32(address, access_type)
        rotate = (address & 3) * 8
        return ror_32(value, rotate)

    def read_16(self, address: int, access_type: MemoryAccess) -> int:
        return self._read_16_internal(address, access_type)

    def read_16_signed(self, address: int, access_type: MemoryAccess) -> int:
        if get_bit(address, 0):
            return self.read_8_signed(address, access_type)
        return extend_sign_16(self.read_16(address, access_type))

    def read_16_ror(self, address: int, access_type: MemoryAccess) -> int:
        value = self.read_16(address, access_type)
        rotate = (address & 1) * 8
        return ror_32(value, rotate)

    def read_8(self, address: int, access_type: MemoryAccess) -> int:
        return self._read_8_internal(address, access_type)

    def read_8_signed(self, address: int, access_type: MemoryAccess) -> int:
        value = self.read_8(address, access_type)
        return extend_sign_8(value)

    def write_32(self, address: int, value: int, access_type: MemoryAccess):
        self._write_32_internal(address, value, access_type)

    def write_16(self, address: int, value: int, access_type: MemoryAccess):
        self._write_16_internal(address, value, access_type)

    def write_8(self, address: int, value: int, access_type: MemoryAccess):
        self._write_8_internal(address, value, access_type)

    def _read_32_internal(self, address: int, access_type: MemoryAccess) -> int:
        self._idle_for_access(address, 4, access_type)
        address = address & ~3
        region = address >> 24
        if region == MemoryRegion.BIOS_REGION:
            if address <= MemoryRegion.BIOS_END:
                if self.cpu.regs.get_pc() <= MemoryRegion.BIOS_END:
                    self.bios_last_opcode = array_read_32(self.bios, address & MemoryRegion.BIOS_MASK)
                return self.bios_last_opcode
            return self._read_unused_memory()
        elif region == MemoryRegion.EWRAM_REGION:
            return array_read_32(self.ewram, address & MemoryRegion.EWRAM_MASK)
        elif region == MemoryRegion.IWRAM_REGION:
            return array_read_32(self.iwram, address & MemoryRegion.IWRAM_MASK)
        elif MemoryRegion.IO_START <= address <= MemoryRegion.IO_END:
            return self.io.read_32(address)
        elif region == MemoryRegion.PALRAM_REGION:
            return self.io.ppu.memory.read_32_palram(address)
        elif region == MemoryRegion.VRAM_REGION:
            return self.io.ppu.memory.read_32_vram(address)
        elif region == MemoryRegion.OAM_REGION:
            return self.io.ppu.memory.read_32_oam(address)
        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            return self.gamepak.read_32(address)
        elif region == MemoryRegion.SRAM_REGION:
            print(f'Attempt to read SRAM: {address:#010x}')
            return 0
        else:
            return self._read_unused_memory()

    def _read_16_internal(self, address: int, access_type: MemoryAccess) -> int:
        self._idle_for_access(address, 2, access_type)
        address = address & ~1
        region = address >> 24
        if region == MemoryRegion.BIOS_REGION:
            if address <= MemoryRegion.BIOS_END:
                if self.cpu.regs.get_pc() <= MemoryRegion.BIOS_END:
                    self.bios_last_opcode = array_read_32(self.bios, address & ~3 & MemoryRegion.BIOS_MASK)
                return self.bios_last_opcode >> (address & 3) * 8 & 65535
            return self._read_unused_memory() >> (address & 3) * 8 & 65535
        elif region == MemoryRegion.EWRAM_REGION:
            return array_read_16(self.ewram, address & MemoryRegion.EWRAM_MASK)
        elif region == MemoryRegion.IWRAM_REGION:
            return array_read_16(self.iwram, address & MemoryRegion.IWRAM_MASK)
        elif MemoryRegion.IO_START <= address <= MemoryRegion.IO_END:
            return self.io.read_16(address)
        elif region == MemoryRegion.PALRAM_REGION:
            return self.io.ppu.memory.read_16_palram(address)
        elif region == MemoryRegion.VRAM_REGION:
            return self.io.ppu.memory.read_16_vram(address)
        elif region == MemoryRegion.OAM_REGION:
            return self.io.ppu.memory.read_16_oam(address)
        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            return self.gamepak.read_16(address)
        elif region == MemoryRegion.SRAM_REGION:
            print(f'Attempt to read SRAM: {address:#010x}')
            return 0
        else:
            return self._read_unused_memory() >> (address & 3) * 8 & 65535

    def _read_8_internal(self, address: int, access_type: MemoryAccess) -> int:
        self._idle_for_access(address, 1, access_type)
        region = address >> 24
        if region == MemoryRegion.BIOS_REGION:
            if address <= MemoryRegion.BIOS_END:
                if self.cpu.regs.get_pc() <= MemoryRegion.BIOS_END:
                    self.bios_last_opcode = array_read_32(self.bios, address & ~3 & MemoryRegion.BIOS_MASK)
                return self.bios_last_opcode >> (address & 3) * 8 & 255
            return self._read_unused_memory() >> (address & 3) * 8 & 255
        elif region == MemoryRegion.EWRAM_REGION:
            return self.ewram[address & MemoryRegion.EWRAM_MASK]
        elif region == MemoryRegion.IWRAM_REGION:
            return self.iwram[address & MemoryRegion.IWRAM_MASK]
        elif MemoryRegion.IO_START <= address <= MemoryRegion.IO_END:
            return self.io.read_8(address)
        elif region == MemoryRegion.PALRAM_REGION:
            return self.io.ppu.memory.read_8_palram(address)
        elif region == MemoryRegion.VRAM_REGION:
            return self.io.ppu.memory.read_8_vram(address)
        elif region == MemoryRegion.OAM_REGION:
            return self.io.ppu.memory.read_8_oam(address)
        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            return self.gamepak.read_8(address)
        elif region == MemoryRegion.SRAM_REGION:
            print(f'Attempt to read SRAM: {address:#010x}')
            return 0
        else:
            return self._read_unused_memory() >> (address & 3) * 8 & 255

    def _write_32_internal(self, address: int, value: int, access_type: MemoryAccess):
        self._idle_for_access(address, 4, access_type)
        address = address & ~3
        region = address >> 24
        if region == MemoryRegion.BIOS_REGION:
            pass
        elif region == MemoryRegion.EWRAM_REGION:
            array_write_32(self.ewram, address & MemoryRegion.EWRAM_MASK, value)
        elif region == MemoryRegion.IWRAM_REGION:
            array_write_32(self.iwram, address & MemoryRegion.IWRAM_MASK, value)
        elif region == MemoryRegion.IO_REGION:
            self.io.write_32(address, value)
        elif region == MemoryRegion.PALRAM_REGION:
            self.io.ppu.memory.write_32_palram(address, value)
        elif region == MemoryRegion.VRAM_REGION:
            self.io.ppu.memory.write_32_vram(address, value)
        elif region == MemoryRegion.OAM_REGION:
            self.io.ppu.memory.write_32_oam(address, value)
        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            print(f'Attempt to write to SRAM: {address:#010x}')
        elif region == MemoryRegion.SRAM_REGION:
            print(f'Attempt to write to SRAM: {address:#010x}')
        else:
            print(f'Attempt to write to unused memory: {address:#010x}')

    def _write_16_internal(self, address: int, value: int, access_type: MemoryAccess):
        self._idle_for_access(address, 2, access_type)
        address = address & ~1
        value = value & 65535
        region = address >> 24
        if region == MemoryRegion.BIOS_REGION:
            pass
        elif region == MemoryRegion.EWRAM_REGION:
            array_write_16(self.ewram, address & MemoryRegion.EWRAM_MASK, value)
        elif region == MemoryRegion.IWRAM_REGION:
            array_write_16(self.iwram, address & MemoryRegion.IWRAM_MASK, value)
        elif region == MemoryRegion.IO_REGION:
            self.io.write_16(address, value)
        elif region == MemoryRegion.PALRAM_REGION:
            self.io.ppu.memory.write_16_palram(address, value)
        elif region == MemoryRegion.VRAM_REGION:
            self.io.ppu.memory.write_16_vram(address, value)
        elif region == MemoryRegion.OAM_REGION:
            self.io.ppu.memory.write_16_oam(address, value)
        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            print(f'Attempt to write to SRAM: {address:#010x}')
        elif region == MemoryRegion.SRAM_REGION:
            print(f'Attempt to write to SRAM: {address:#010x}')
        else:
            print(f'Attempt to write to unused memory: {address:#010x}')

    def _write_8_internal(self, address: int, value: int, access_type: MemoryAccess):
        self._idle_for_access(address, 1, access_type)
        value = value & 255
        region = address >> 24
        if region == MemoryRegion.BIOS_REGION:
            pass
        elif region == MemoryRegion.EWRAM_REGION:
            self.ewram[address & MemoryRegion.EWRAM_MASK] = value
        elif region == MemoryRegion.IWRAM_REGION:
            self.iwram[address & MemoryRegion.IWRAM_MASK] = value
        elif region == MemoryRegion.IO_REGION:
            self.io.write_8(address, value)
        elif region == MemoryRegion.PALRAM_REGION:
            self.io.ppu.memory.write_8_palram(address, value)
        elif region == MemoryRegion.VRAM_REGION:
            self.io.ppu.memory.write_8_vram(address, value)
        elif region == MemoryRegion.OAM_REGION:
            pass
        elif MemoryRegion.GAMEPAK_REGION_START <= region <= MemoryRegion.GAMEPAK_REGION_END:
            print(f'Attempt to write to SRAM: {address:#010x}')
        elif region == MemoryRegion.SRAM_REGION:
            print(f'Attempt to write to SRAM: {address:#010x}')
        else:
            print(f'Attempt to write to unused memory: {address:#010x}')

    def _read_unused_memory(self) -> int:
        """
        Reading from unused memory returns the last prefetched instruction.
        https://problemkaputt.de/gbatek.htm#gbaunpredictablethings

        For THUMB code, the result consists of two 16-bit fragments and depends on
        the address area and alignment where the opcode was stored.
        """
        if self.cpu.regs.cpsr.get_state() == CPUState.ARM:
            return self.cpu.pipeline[1]
        region = self.cpu.regs.get_pc() >> 24
        if region == MemoryRegion.BIOS_REGION or region == MemoryRegion.OAM_REGION:
            if self.cpu.regs.get_pc() & 3 == 0:
                return self.cpu.pipeline[1] << 16 | self.cpu.pipeline[1]
            else:
                return self.cpu.pipeline[1] << 16 | self.cpu.pipeline[0]
        elif region == MemoryRegion.IWRAM_REGION:
            if self.cpu.regs.get_pc() & 3 == 0:
                return self.cpu.pipeline[0] << 16 | self.cpu.pipeline[1]
            else:
                return self.cpu.pipeline[1] << 16 | self.cpu.pipeline[0]
        else:
            return self.cpu.pipeline[1] << 16 | self.cpu.pipeline[1]

    def _init_access_times(self):
        """
        Source: GBATEK

        Region        Bus   Read      Write     Cycles
        =====================================================
        BIOS ROM      32    8/16/32   -         1/1/1
        Work RAM 32K  32    8/16/32   8/16/32   1/1/1
        I/O           32    8/16/32   8/16/32   1/1/1
        OAM           32    8/16/32   16/32     1/1/1 *
        Work RAM 256K 16    8/16/32   8/16/32   3/3/6 **
        Palette RAM   16    8/16/32   16/32     1/1/2 *
        VRAM          16    8/16/32   16/32     1/1/2 *
        GamePak ROM   16    8/16/32   -         5/5/8 **/***
        GamePak Flash 16    8/16/32   16/32     5/5/8 **/***
        GamePak SRAM  8     8         8         5     **

        Timing Notes:

          *   Plus 1 cycle if GBA accesses video memory at the same time.
          **  Default waitstate settings, see System Control chapter.
          *** Separate timings for sequential, and non-sequential accesses.

        """
        self.access_time_32 = [[1] * 16 for _ in range(2)]
        self.access_time_16 = [[1] * 16 for _ in range(2)]
        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)] = 6
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)] = 6
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)] = 3
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.EWRAM_REGION)] = 3
        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.PALRAM_REGION)] = 2
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.PALRAM_REGION)] = 2
        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.VRAM_REGION)] = 2
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.VRAM_REGION)] = 2
        self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.OAM_REGION)] = 2
        self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.OAM_REGION)] = 2
        self.update_waitstates()

    def update_waitstates(self):
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_1)] = 1 + self.wait_control.get_ws0_non_seq()
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_2)] = 1 + self.wait_control.get_ws0_non_seq()
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_1)] = 1 + self.wait_control.get_ws1_non_seq()
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_2)] = 1 + self.wait_control.get_ws1_non_seq()
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_1)] = 1 + self.wait_control.get_ws2_non_seq()
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_2)] = 1 + self.wait_control.get_ws2_non_seq()
        self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(MemoryRegion.SRAM_REGION)] = 1 + self.wait_control.get_sram()
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_1)] = 1 + self.wait_control.get_ws0_seq()
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_0_REGION_2)] = 1 + self.wait_control.get_ws0_seq()
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_1)] = 1 + self.wait_control.get_ws1_seq()
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_1_REGION_2)] = 1 + self.wait_control.get_ws1_seq()
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_1)] = 1 + self.wait_control.get_ws2_seq()
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.GAMEPAK_2_REGION_2)] = 1 + self.wait_control.get_ws2_seq()
        self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(MemoryRegion.SRAM_REGION)] = 1 + self.wait_control.get_sram()
        for region in range(MemoryRegion.GAMEPAK_0_REGION_1, MemoryRegion.SRAM_REGION + 1):
            self.access_time_32[int(MemoryAccess.NON_SEQUENTIAL)][int(region)] = self.access_time_16[int(MemoryAccess.NON_SEQUENTIAL)][int(region)] + self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(region)]
            self.access_time_32[int(MemoryAccess.SEQUENTIAL)][int(region)] = self.access_time_16[int(MemoryAccess.SEQUENTIAL)][int(region)] * 2

    def _idle_for_access(self, address: int, size: int, access_type: MemoryAccess):
        region = address >> 24 & 15
        cycles = self.access_time_32[int(access_type)][int(region)] if size == 4 else self.access_time_16[int(access_type)][int(region)]
        self.scheduler.idle(cycles)

class WaitstateControlRegister:

    def __init__(self):
        self.NON_SEQUENTIAL_CYCLES = [4, 3, 2, 8]
        self.reg = 0

    def get_sram(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 0, 1)]

    def get_ws0_non_seq(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 2, 3)]

    def get_ws0_seq(self):
        return 1 if get_bit(self.reg, 4) else 2

    def get_ws1_non_seq(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 5, 6)]

    def get_ws1_seq(self):
        return 1 if get_bit(self.reg, 7) else 4

    def get_ws2_non_seq(self):
        return self.NON_SEQUENTIAL_CYCLES[get_bits(self.reg, 8, 9)]

    def get_ws2_seq(self):
        return 1 if get_bit(self.reg, 10) else 8


from enum import IntFlag, IntEnum
CLOCK_SPEED_HZ = 16777216
NANOSECONDS_PER_SECOND = 1000000000

class Interrupt(IntFlag):
    VBLANK = 1 << 0
    HBLANK = 1 << 1
    VCOUNT = 1 << 2
    TIMER_0 = 1 << 3
    TIMER_1 = 1 << 4
    TIMER_2 = 1 << 5
    TIMER_3 = 1 << 6
    SERIAL = 1 << 7
    DMA_0 = 1 << 8
    DMA_1 = 1 << 9
    DMA_2 = 1 << 10
    DMA_3 = 1 << 11
    KEYPAD = 1 << 12
    GAMEPAK = 1 << 13
    ALL = 16383

class PowerDownMode(IntEnum):
    NONE = 0
    HALT = 1
    STOP = 2

class EventTrigger(IntEnum):
    IMMEDIATELY = 0
    VBLANK = 1
    HBLANK = 2

class Key(IntFlag):
    BUTTON_A = 1 << 0
    BUTTON_B = 1 << 1
    BUTTON_SELECT = 1 << 2
    BUTTON_START = 1 << 3
    DPAD_RIGHT = 1 << 4
    DPAD_LEFT = 1 << 5
    DPAD_UP = 1 << 6
    DPAD_DOWN = 1 << 7
    SHOULDER_RIGHT = 1 << 8
    SHOULDER_LEFT = 1 << 9


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.memory.memory import Memory

from pyboy_advance.interrupt_controller import InterruptController
from pyboy_advance.keypad import Keypad
from pyboy_advance.memory.constants import IOAddress
from pyboy_advance.memory.dma import DMAController
from pyboy_advance.ppu.ppu import PPU
from pyboy_advance.timer import Timers
from pyboy_advance.utils import get_bit
# endif


class IO:
    def __init__(
        self,
        memory: Memory,
        interrupt_controller: InterruptController,
        dma_controller: DMAController,
        timers: Timers,
        ppu: PPU,
        keypad: Keypad,
    ):
        self.memory = memory
        self.interrupt_controller = interrupt_controller
        self.dma_controller = dma_controller
        self.timers = timers
        self.ppu = ppu
        self.keypad = keypad
        self.reg_soundbias = 0

    def read_32(self, address: int) -> int:
        lower_bits = self.read_16(address)
        upper_bits = self.read_16(address + 2)
        return (upper_bits << 16) | lower_bits

    def read_16(self, address: int) -> int:
        # Display Control Registers
        if address == IOAddress.REG_DISPCNT:
            return self.ppu.display_control.reg
        elif address == IOAddress.REG_DISPSTAT:
            return self.ppu.display_status.reg
        elif address == IOAddress.REG_VCOUNT:
            return self.ppu.vcount

        # Background Control Registers
        elif address == IOAddress.REG_BG0CNT:
            return self.ppu.bg_control_0.reg
        elif address == IOAddress.REG_BG1CNT:
            return self.ppu.bg_control_1.reg
        elif address == IOAddress.REG_BG2CNT:
            return self.ppu.bg_control_2.reg
        elif address == IOAddress.REG_BG3CNT:
            return self.ppu.bg_control_3.reg

        # Window Registers
        elif address == IOAddress.REG_WININ:
            win_in = self.ppu.window_control_0.reg
            win_in |= self.ppu.window_control_1.reg << 8
            return win_in
        elif address == IOAddress.REG_WINOUT:
            win_out = self.ppu.window_control_out.reg
            win_out |= self.ppu.window_control_obj.reg << 8
            return win_out

        # Special Effects Registers
        elif address == IOAddress.REG_BLDCNT:
            return self.ppu.blend_control.reg
        elif address == IOAddress.REG_BLDALPHA:
            return self.ppu.blend_alpha.reg

        # Sound Control Registers
        elif address == IOAddress.REG_SOUNDBIAS:
            return self.reg_soundbias

        # DMA Control Registers
        elif address == IOAddress.REG_DMA0CNT_H:
            return self.dma_controller.channel_0.control_reg
        elif address == IOAddress.REG_DMA1CNT_H:
            return self.dma_controller.channel_1.control_reg
        elif address == IOAddress.REG_DMA2CNT_H:
            return self.dma_controller.channel_2.control_reg
        elif address == IOAddress.REG_DMA3CNT_H:
            return self.dma_controller.channel_3.control_reg

        # Timer Registers
        elif address == IOAddress.REG_TM0CNT_L:
            return self.timers.timer_0.counter
        elif address == IOAddress.REG_TM0CNT_H:
            return self.timers.timer_0.control_reg
        elif address == IOAddress.REG_TM1CNT_L:
            return self.timers.timer_1.counter
        elif address == IOAddress.REG_TM1CNT_H:
            return self.timers.timer_1.control_reg
        elif address == IOAddress.REG_TM2CNT_L:
            return self.timers.timer_2.counter
        elif address == IOAddress.REG_TM2CNT_H:
            return self.timers.timer_2.control_reg
        elif address == IOAddress.REG_TM3CNT_L:
            return self.timers.timer_3.counter
        elif address == IOAddress.REG_TM3CNT_H:
            return self.timers.timer_3.control_reg

        # Keypad Registers
        elif address == IOAddress.REG_KEYINPUT:
            return self.keypad.key_input
        elif address == IOAddress.REG_KEYCNT:
            return self.keypad.key_control.reg

        # Interrupt Registers
        elif address == IOAddress.REG_IE:
            return self.interrupt_controller.interrupt_enable
        elif address == IOAddress.REG_IF:
            return self.interrupt_controller.interrupt_flags
        elif address == IOAddress.REG_WAITCNT:
            return self.memory.wait_control.reg
        elif address == IOAddress.REG_IME:
            return self.interrupt_controller.interrupt_master_enable

        else:
            return 0

    def read_8(self, address: int) -> int:
        value = self.read_16(address & ~1)
        return (value >> ((address & 1) * 8)) & 0xFF

    def write_32(self, address: int, value: int):
        self.write_16(address, value & 0xFFFF)
        self.write_16(address + 2, (value >> 16) & 0xFFFF)

    def write_16(self, address: int, value: int):
        # Display Control Registers
        if address == IOAddress.REG_DISPCNT:
            self.ppu.display_control.reg = value
        elif address == IOAddress.REG_DISPSTAT:
            self.ppu.display_status.reg = value
        elif address == IOAddress.REG_VCOUNT:
            self.ppu.vcount = value

        # Background Control Registers
        elif address == IOAddress.REG_BG0CNT:
            self.ppu.bg_control_0.reg = value
        elif address == IOAddress.REG_BG1CNT:
            self.ppu.bg_control_1.reg = value
        elif address == IOAddress.REG_BG2CNT:
            self.ppu.bg_control_2.reg = value
        elif address == IOAddress.REG_BG3CNT:
            self.ppu.bg_control_3.reg = value

        # Background Offset Registers
        elif address == IOAddress.REG_BG0HOFS:
            self.ppu.bg_offset_h[0] = value & 0x3FF
        elif address == IOAddress.REG_BG0VOFS:
            self.ppu.bg_offset_v[0] = value & 0x3FF
        elif address == IOAddress.REG_BG1HOFS:
            self.ppu.bg_offset_h[1] = value & 0x3FF
        elif address == IOAddress.REG_BG1VOFS:
            self.ppu.bg_offset_v[1] = value & 0x3FF
        elif address == IOAddress.REG_BG2HOFS:
            self.ppu.bg_offset_h[2] = value & 0x3FF
        elif address == IOAddress.REG_BG2VOFS:
            self.ppu.bg_offset_v[2] = value & 0x3FF
        elif address == IOAddress.REG_BG3HOFS:
            self.ppu.bg_offset_h[3] = value & 0x3FF
        elif address == IOAddress.REG_BG3VOFS:
            self.ppu.bg_offset_v[3] = value & 0x3FF

        # Window Registers
        elif address == IOAddress.REG_WIN0H:
            self.ppu.window_h_min[0] = (value >> 8) & 0xFF
            self.ppu.window_h_max[0] = value & 0xFF
        elif address == IOAddress.REG_WIN1H:
            self.ppu.window_h_min[1] = (value >> 8) & 0xFF
            self.ppu.window_h_max[1] = value & 0xFF
        elif address == IOAddress.REG_WIN0V:
            self.ppu.window_v_min[0] = (value >> 8) & 0xFF
            self.ppu.window_v_max[0] = value & 0xFF
        elif address == IOAddress.REG_WIN1V:
            self.ppu.window_v_min[1] = (value >> 8) & 0xFF
            self.ppu.window_v_max[1] = value & 0xFF
        elif address == IOAddress.REG_WININ:
            self.ppu.window_control_0.reg = value & 0x3F
            self.ppu.window_control_1.reg = (value >> 8) & 0x3F
        elif address == IOAddress.REG_WINOUT:
            self.ppu.window_control_out.reg = value & 0x3F
            self.ppu.window_control_obj.reg = (value >> 8) & 0x3F

        # Special Effects Registers
        elif address == IOAddress.REG_BLDCNT:
            self.ppu.blend_control.reg = value
        elif address == IOAddress.REG_BLDALPHA:
            self.ppu.blend_alpha.reg = value
        elif address == IOAddress.REG_BLDY:
            self.ppu.blend_brightness.reg = value

        # Sound Control Registers
        elif address == IOAddress.REG_SOUNDBIAS:
            self.reg_soundbias = value

        # DMA Source/Destination Registers
        elif address == IOAddress.REG_DMA0SAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_0.src_address &= mask
            self.dma_controller.channel_0.src_address |= value
        elif address == IOAddress.REG_DMA0SAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_0.src_address &= mask
            self.dma_controller.channel_0.src_address |= value << 16
        elif address == IOAddress.REG_DMA0DAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_0.dst_address &= mask
            self.dma_controller.channel_0.dst_address |= value
        elif address == IOAddress.REG_DMA0DAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_0.dst_address &= mask
            self.dma_controller.channel_0.dst_address |= value << 16
        elif address == IOAddress.REG_DMA1SAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_1.src_address &= mask
            self.dma_controller.channel_1.src_address |= value
        elif address == IOAddress.REG_DMA1SAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_1.src_address &= mask
            self.dma_controller.channel_1.src_address |= value << 16
        elif address == IOAddress.REG_DMA1DAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_1.dst_address &= mask
            self.dma_controller.channel_1.dst_address |= value
        elif address == IOAddress.REG_DMA1DAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_1.dst_address &= mask
            self.dma_controller.channel_1.dst_address |= value << 16
        elif address == IOAddress.REG_DMA2SAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_2.src_address &= mask
            self.dma_controller.channel_2.src_address |= value
        elif address == IOAddress.REG_DMA2SAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_2.src_address &= mask
            self.dma_controller.channel_2.src_address |= value << 16
        elif address == IOAddress.REG_DMA2DAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_2.dst_address &= mask
            self.dma_controller.channel_2.dst_address |= value
        elif address == IOAddress.REG_DMA2DAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_2.dst_address &= mask
            self.dma_controller.channel_2.dst_address |= value << 16
        elif address == IOAddress.REG_DMA3SAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_3.src_address &= mask
            self.dma_controller.channel_3.src_address |= value
        elif address == IOAddress.REG_DMA3SAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_3.src_address &= mask
            self.dma_controller.channel_3.src_address |= value << 16
        elif address == IOAddress.REG_DMA3DAD_L:
            mask = 0xFFFF0000
            self.dma_controller.channel_3.dst_address &= mask
            self.dma_controller.channel_3.dst_address |= value
        elif address == IOAddress.REG_DMA3DAD_H:
            mask = 0xFFFF
            self.dma_controller.channel_3.dst_address &= mask
            self.dma_controller.channel_3.dst_address |= value << 16

        # DMA Control Registers
        elif address == IOAddress.REG_DMA0CNT_L:
            self.dma_controller.channel_0.count = value
        elif address == IOAddress.REG_DMA0CNT_H:
            self.dma_controller.channel_0.control_reg = value
        elif address == IOAddress.REG_DMA1CNT_L:
            self.dma_controller.channel_1.count = value
        elif address == IOAddress.REG_DMA1CNT_H:
            self.dma_controller.channel_1.control_reg = value
        elif address == IOAddress.REG_DMA2CNT_L:
            self.dma_controller.channel_2.count = value
        elif address == IOAddress.REG_DMA2CNT_H:
            self.dma_controller.channel_2.control_reg = value
        elif address == IOAddress.REG_DMA3CNT_L:
            self.dma_controller.channel_3.count = value
        elif address == IOAddress.REG_DMA3CNT_H:
            self.dma_controller.channel_3.control_reg = value

        # Timer Registers
        elif address == IOAddress.REG_TM0CNT_L:
            self.timers.timer_0.counter = value
        elif address == IOAddress.REG_TM0CNT_H:
            self.timers.timer_0.control_reg = value
        elif address == IOAddress.REG_TM1CNT_L:
            self.timers.timer_1.counter = value
        elif address == IOAddress.REG_TM1CNT_H:
            self.timers.timer_1.control_reg = value
        elif address == IOAddress.REG_TM2CNT_L:
            self.timers.timer_2.counter = value
        elif address == IOAddress.REG_TM2CNT_H:
            self.timers.timer_2.control_reg = value
        elif address == IOAddress.REG_TM3CNT_L:
            self.timers.timer_3.counter = value
        elif address == IOAddress.REG_TM3CNT_H:
            self.timers.timer_3.control_reg = value

        # Keypad Registers
        elif address == IOAddress.REG_KEYCNT:
            self.keypad.key_control.reg = value

        # Interrupt Registers
        elif address == IOAddress.REG_IE:
            self.interrupt_controller.interrupt_enable = value
        elif address == IOAddress.REG_IF:
            self.interrupt_controller.interrupt_flags = value
        elif address == IOAddress.REG_WAITCNT:
            self.memory.wait_control.reg = value
            self.memory.update_waitstates()
        elif address == IOAddress.REG_IME:
            self.interrupt_controller.interrupt_master_enable = value
        elif address == IOAddress.REG_HALTCNT:
            self.interrupt_controller.power_down_mode = get_bit(value, 15) + 1

    def write_8(self, address: int, value: int):
        aligned_address = address & ~1

        old_value = self.read_16(aligned_address)
        new_value = (
            (old_value & 0x00FF) | (value << 8)
            if get_bit(address, 0)
            else (old_value & 0xFF00) | value
        )

        self.write_16(aligned_address, new_value)



from enum import IntEnum

class WindowEvent(IntEnum):
    NONE = 1
    QUIT = 2
    FULLSCREEN = 3
    PRESS_BUTTON_A = 4
    PRESS_BUTTON_B = 5
    PRESS_BUTTON_SELECT = 6
    PRESS_BUTTON_START = 7
    PRESS_DPAD_RIGHT = 8
    PRESS_DPAD_LEFT = 9
    PRESS_DPAD_UP = 10
    PRESS_DPAD_DOWN = 11
    PRESS_SHOULDER_RIGHT = 12
    PRESS_SHOULDER_LEFT = 13
    RELEASE_BUTTON_A = 14
    RELEASE_BUTTON_B = 15
    RELEASE_BUTTON_SELECT = 16
    RELEASE_BUTTON_START = 17
    RELEASE_DPAD_RIGHT = 18
    RELEASE_DPAD_LEFT = 19
    RELEASE_DPAD_UP = 20
    RELEASE_DPAD_DOWN = 21
    RELEASE_SHOULDER_RIGHT = 22
    RELEASE_SHOULDER_LEFT = 23


cpdef enum ALUOpcode:
    AND = 0
    EOR = 1
    SUB = 2
    RSB = 3
    ADD = 4
    ADC = 5
    SBC = 6
    RSC = 7
    TST = 8
    TEQ = 9
    CMP = 10
    CMN = 11
    ORR = 12
    MOV = 13
    BIC = 14
    MVN = 15

cpdef enum DataTransferOpcode:
    STRH = 1
    LDRD = 2
    STRD = 3
    LDRH = 1
    LDRSB = 2
    LDRSH = 3

cpdef enum MultiplyLongOpcode:
    UMULL = 0
    UMLAL = 1
    SMULL = 2
    SMLAL = 3




cimport cython

from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.constants cimport CPUMode
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport get_bits, get_bit, ror_32

cdef void arm_mrs(CPU, uint32_t) noexcept

@cython.locals(mask=uint32_t, flags_mask=uint32_t, status_mask=uint32_t, extension_mask=uint32_t, control_mask=uint32_t)
cdef void arm_msr(CPU, uint32_t) noexcept



import pytest


@pytest.mark.rom_url("https://raw.githubusercontent.com/DenSinH/FuzzARM/master/FuzzARM.gba")
def test_fuzzarm(emulator, assert_expected_image):
    emulator.frame(1000)
    assert_expected_image(emulator.screen.image)



# ifndef CYTHON
from __future__ import annotations

from array import array

from pyboy_advance.cpu.constants import CPUMode, CPUState, BankIndex
from pyboy_advance.utils import get_bit, set_bit, bint
# endif


class Registers:
    """
    ARM7TDMI register set. All registers are 32 bits wide.

    Only 16 registers are visible at any given time.
    There are 20 "banked" registers that get swapped in whenever the CPU
    switches to various privileged modes.
    """

    def __init__(self):
        self.SP = 13
        self.LR = 14
        self.PC = 15

        # General purpose registers R8 to R12 are banked
        self.BANKED_GPR_RANGE_START = 8
        self.BANKED_GPR_RANGE_END = 12
        self.BANKED_GPR_RANGE_LEN = self.BANKED_GPR_RANGE_END - self.BANKED_GPR_RANGE_START + 1

        self.regs = array("I", [0] * 16)
        self.cpsr = ProgramStatusRegister()  # Current program status register
        self.spsr = ProgramStatusRegister()  # Saved program status register

        # System/User mode and FIQ mode each have their own copy of R8 to R12
        self.banked_old_gpr = array("I", [0] * self.BANKED_GPR_RANGE_LEN)
        self.banked_fiq_gpr = array("I", [0] * self.BANKED_GPR_RANGE_LEN)

        # Every mode has its own SP and LR register
        self.banked_sp = array("I", [0] * len(BankIndex))
        self.banked_lr = array("I", [0] * len(BankIndex))

        # Every mode has its own SPSR register
        self.banked_spsr_user = ProgramStatusRegister()
        self.banked_spsr_fiq = ProgramStatusRegister()
        self.banked_spsr_irq = ProgramStatusRegister()
        self.banked_spsr_swi = ProgramStatusRegister()
        self.banked_spsr_abort = ProgramStatusRegister()
        self.banked_spsr_undefined = ProgramStatusRegister()

    def get(self, reg: int):
        return self.regs[reg]

    def set(self, reg: int, value: int):
        self.regs[reg] = value

    @property
    def sp(self):
        return self.regs[self.SP]

    @sp.setter
    def sp(self, value: int):
        self.regs[self.SP] = value

    @property
    def lr(self):
        return self.regs[self.LR]

    @lr.setter
    def lr(self, value: int):
        self.regs[self.LR] = value

    @property
    def pc(self):
        return self.regs[self.PC]

    @pc.setter
    def pc(self, value: int):
        self.regs[self.PC] = value

    def switch_mode(self, new_mode: CPUMode):
        """
        Switch from the current mode to the specified mode.

        Saves the current register values to the current mode's bank and
        replaces them with values from the new mode's bank.
        """
        if new_mode == self.cpsr.mode:
            return

        old_mode = self.cpsr.mode
        self.cpsr.mode = new_mode

        old_bank_index = self.get_bank_index(old_mode)
        new_bank_index = self.get_bank_index(new_mode)

        self.banked_sp[old_bank_index] = self.sp
        self.banked_lr[old_bank_index] = self.lr
        self.get_banked_spsr(old_bank_index).reg = self.spsr.reg

        self.sp = self.banked_sp[new_bank_index]
        self.lr = self.banked_lr[new_bank_index]
        self.spsr.reg = self.get_banked_spsr(new_bank_index).reg

        if new_mode == CPUMode.FIQ:
            for i in range(self.BANKED_GPR_RANGE_LEN):
                self.banked_old_gpr[i] = self.regs[i + self.BANKED_GPR_RANGE_START]
                self.regs[i + self.BANKED_GPR_RANGE_START] = self.banked_fiq_gpr[i]
        elif old_mode == CPUMode.FIQ:
            for i in range(self.BANKED_GPR_RANGE_LEN):
                self.banked_fiq_gpr[i] = self.regs[i + self.BANKED_GPR_RANGE_START]
                self.regs[i + self.BANKED_GPR_RANGE_START] = self.banked_old_gpr[i]

    def get_bank_index(self, mode: CPUMode) -> BankIndex:
        if mode == CPUMode.SYSTEM or mode == CPUMode.USER:
            return BankIndex.BANK_SYSTEM_USER
        elif mode == CPUMode.FIQ:
            return BankIndex.BANK_FIQ
        elif mode == CPUMode.IRQ:
            return BankIndex.BANK_IRQ
        elif mode == CPUMode.SWI:
            return BankIndex.BANK_SWI
        elif mode == CPUMode.ABORT:
            return BankIndex.BANK_ABORT
        elif mode == CPUMode.UNDEFINED:
            return BankIndex.BANK_UNDEFINED
        # ifndef CYTHON
        raise ValueError("Invalid CPU mode")
        # endif

    def get_banked_spsr(self, bank_index: BankIndex) -> ProgramStatusRegister:
        if bank_index == BankIndex.BANK_SYSTEM_USER:
            return self.banked_spsr_user
        if bank_index == BankIndex.BANK_FIQ:
            return self.banked_spsr_fiq
        elif bank_index == BankIndex.BANK_IRQ:
            return self.banked_spsr_irq
        elif bank_index == BankIndex.BANK_SWI:
            return self.banked_spsr_swi
        elif bank_index == BankIndex.BANK_ABORT:
            return self.banked_spsr_abort
        elif bank_index == BankIndex.BANK_UNDEFINED:
            return self.banked_spsr_undefined
        # ifndef CYTHON
        raise ValueError("Invalid bank index")
        # endif


class ProgramStatusRegister:
    def __init__(self, value: int = 0):
        self.reg = value

    @property
    def mode(self) -> CPUMode | int:
        return self.reg & 0b11111

    @mode.setter
    def mode(self, mode: CPUMode):
        self.reg = (self.reg & ~0b11111) | mode

    @property
    def state(self) -> CPUState | int:
        return get_bit(self.reg, 5)

    @state.setter
    def state(self, state: CPUState):
        self.reg = set_bit(self.reg, 5, state)

    @property
    def sign_flag(self) -> bint:
        return get_bit(self.reg, 31)

    @sign_flag.setter
    def sign_flag(self, value: bint):
        self.reg = set_bit(self.reg, 31, value)

    @property
    def zero_flag(self) -> bint:
        return get_bit(self.reg, 30)

    @zero_flag.setter
    def zero_flag(self, value: bint):
        self.reg = set_bit(self.reg, 30, value)

    @property
    def carry_flag(self) -> bint:
        return get_bit(self.reg, 29)

    @carry_flag.setter
    def carry_flag(self, value: bint):
        self.reg = set_bit(self.reg, 29, value)

    @property
    def overflow_flag(self) -> bint:
        return get_bit(self.reg, 28)

    @overflow_flag.setter
    def overflow_flag(self, value: bint):
        self.reg = set_bit(self.reg, 28, value)

    @property
    def sticky_overflow_flag(self) -> bint:
        return get_bit(self.reg, 27)

    @sticky_overflow_flag.setter
    def sticky_overflow_flag(self, value: bint):
        self.reg = set_bit(self.reg, 27, value)

    @property
    def irq_disable(self) -> bint:
        return get_bit(self.reg, 7)

    @irq_disable.setter
    def irq_disable(self, value: bint):
        self.reg = set_bit(self.reg, 7, value)

    @property
    def fiq_disable(self) -> bint:
        return get_bit(self.reg, 6)

    @fiq_disable.setter
    def fiq_disable(self, value: bint):
        self.reg = set_bit(self.reg, 6, value)



import pytest


@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/arm/arm.gba")
def test_jsmolka_arm(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/bios/bios.gba")
def test_jsmolka_bios(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/memory/memory.gba")
def test_jsmolka_memory(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/nes/nes.gba")
def test_jsmolka_nes(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/save/flash64.gba")
def test_jsmolka_save_flash64(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/save/flash128.gba")
def test_jsmolka_save_flash128(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/save/none.gba")
def test_jsmolka_save_none(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/save/sram.gba")
def test_jsmolka_save_sram(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/thumb/thumb.gba")
def test_jsmolka_thumb(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://raw.githubusercontent.com/jsmolka/gba-tests/master/unsafe/unsafe.gba")
def test_jsmolka_unsafe(emulator, assert_expected_image):
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)



class CPU:
    """
    ARM7TDMI, a 32-bit RISC processor. It has a 3 stage instruction pipeline
    (fetch, decode, execute).

    The ARM7TDMI can operate in one of two different CPU states:
     - ARM, which uses a full instruction set with 32-bit opcodes
     - THUMB, which uses a cut-down instruction set with 16-bit opcodes

    https://problemkaputt.de/gbatek.htm#armcpureference
    """

    def __init__(self, scheduler: Scheduler, memory: Memory):
        self.scheduler = scheduler
        self.regs = Registers()
        self.regs.cpsr.set_mode(CPUMode.SYSTEM)
        self.regs.spsr.set_mode(CPUMode.SYSTEM)
        self.memory = memory
        self.memory.cpu = self
        self.pipeline = [4026531840, 4026531840]
        self.next_fetch_access = MemoryAccess.NON_SEQUENTIAL

    def step(self):
        irq_line = self.memory.io.interrupt_controller.irq_line
        if irq_line and (not self.regs.cpsr.get_irq_disable()):
            self.interrupt(ExceptionVector.ExceptionVector_IRQ)
        power_down_mode = self.memory.io.interrupt_controller.power_down_mode
        if power_down_mode == PowerDownMode.NONE:
            if self.regs.cpsr.get_state() == CPUState.ARM:
                self.step_arm()
            else:
                self.step_thumb()
        elif power_down_mode == PowerDownMode.HALT:
            self.scheduler.idle_until_next_event()

    def step_arm(self):
        instruction = self.pipeline[0]
        self.pipeline[0] = self.pipeline[1]
        self.pipeline[1] = self.memory.read_32(self.regs.get_pc(), self.next_fetch_access)
        instruction_handler = self.arm_decoder(instruction)
        cond = get_bits(instruction, 28, 31)
        if self.check_condition(cond):
            instruction_handler(self, instruction)
        else:
            self.advance_pc_arm()
            self.next_fetch_access = MemoryAccess.SEQUENTIAL

    def step_thumb(self):
        instruction = self.pipeline[0]
        self.pipeline[0] = self.pipeline[1]
        self.pipeline[1] = self.memory.read_16(self.regs.get_pc(), self.next_fetch_access)
        instruction_handler = self.thumb_decoder(instruction)
        instruction_handler(self, instruction)

    def advance_pc_arm(self):
        self.regs.set_pc(add_32(self.regs.get_pc(), ARM_PC_INCREMENT))

    def advance_pc_thumb(self):
        self.regs.set_pc(add_32(self.regs.get_pc(), THUMB_PC_INCREMENT))

    def flush_pipeline(self):
        if self.regs.cpsr.get_state() == CPUState.ARM:
            self.regs.set_pc(self.regs.get_pc() & ~3)
            self.pipeline[0] = self.memory.read_32(self.regs.get_pc(), MemoryAccess.NON_SEQUENTIAL)
            self.advance_pc_arm()
            self.pipeline[1] = self.memory.read_32(self.regs.get_pc(), MemoryAccess.SEQUENTIAL)
            self.advance_pc_arm()
        else:
            self.regs.set_pc(self.regs.get_pc() & ~1)
            self.pipeline[0] = self.memory.read_16(self.regs.get_pc(), MemoryAccess.NON_SEQUENTIAL)
            self.advance_pc_thumb()
            self.pipeline[1] = self.memory.read_16(self.regs.get_pc(), MemoryAccess.SEQUENTIAL)
            self.advance_pc_thumb()
        self.next_fetch_access = MemoryAccess.SEQUENTIAL

    def switch_mode(self, new_mode: CPUMode):
        self.regs.switch_mode(new_mode)

    def interrupt(self, vector: ExceptionVector):
        if vector == ExceptionVector.RESET:
            new_mode = CPUMode.SWI
        elif vector == ExceptionVector.UNDEFINED_INSTRUCTION:
            new_mode = CPUMode.UNDEFINED
        elif vector == ExceptionVector.ExceptionVector_SWI:
            new_mode = CPUMode.SWI
        elif vector == ExceptionVector.PREFETCH_ABORT:
            new_mode = CPUMode.ABORT
        elif vector == ExceptionVector.DATA_ABORT:
            new_mode = CPUMode.ABORT
        elif vector == ExceptionVector.ADDRESS_EXCEEDS_26_BITS:
            new_mode = CPUMode.SWI
        elif vector == ExceptionVector.ExceptionVector_IRQ:
            new_mode = CPUMode.IRQ
        elif vector == ExceptionVector.ExceptionVector_FIQ:
            new_mode = CPUMode.FIQ
        else:
            return
        cpsr_reg = self.regs.cpsr.reg
        self.switch_mode(new_mode)
        self.regs.spsr.reg = cpsr_reg
        if vector == ExceptionVector.ExceptionVector_SWI or vector == ExceptionVector.UNDEFINED_INSTRUCTION:
            self.regs.set_lr(add_32(self.regs.get_pc(), -4 if self.regs.cpsr.get_state() == CPUState.ARM else -2))
        elif vector != ExceptionVector.RESET:
            self.regs.set_lr(add_32(self.regs.get_pc(), -4 if self.regs.cpsr.get_state() == CPUState.ARM else 0))
        self.regs.cpsr.set_state(CPUState.ARM)
        self.regs.cpsr.set_irq_disable(True)
        if vector in [ExceptionVector.RESET, ExceptionVector.ExceptionVector_FIQ]:
            self.regs.cpsr.set_fiq_disable(True)
        self.regs.set_pc(vector)
        self.flush_pipeline()

    def check_condition(self, cond: Condition | int) -> bint:
        if cond == Condition.EQ:
            return self.regs.cpsr.get_zero_flag()
        elif cond == Condition.NE:
            return not self.regs.cpsr.get_zero_flag()
        elif cond == Condition.HS:
            return self.regs.cpsr.get_carry_flag()
        elif cond == Condition.LO:
            return not self.regs.cpsr.get_carry_flag()
        elif cond == Condition.MI:
            return self.regs.cpsr.get_sign_flag()
        elif cond == Condition.PL:
            return not self.regs.cpsr.get_sign_flag()
        elif cond == Condition.VS:
            return self.regs.cpsr.get_overflow_flag()
        elif cond == Condition.VC:
            return not self.regs.cpsr.get_overflow_flag()
        elif cond == Condition.HI:
            return self.regs.cpsr.get_carry_flag() and (not self.regs.cpsr.get_zero_flag())
        elif cond == Condition.LS:
            return not self.regs.cpsr.get_carry_flag() or self.regs.cpsr.get_zero_flag()
        elif cond == Condition.GE:
            return self.regs.cpsr.get_sign_flag() == self.regs.cpsr.get_overflow_flag()
        elif cond == Condition.LT:
            return self.regs.cpsr.get_sign_flag() != self.regs.cpsr.get_overflow_flag()
        elif cond == Condition.GT:
            return not self.regs.cpsr.get_zero_flag() and self.regs.cpsr.get_sign_flag() == self.regs.cpsr.get_overflow_flag()
        elif cond == Condition.LE:
            return self.regs.cpsr.get_zero_flag() or self.regs.cpsr.get_sign_flag() != self.regs.cpsr.get_overflow_flag()
        elif cond == Condition.AL:
            return True
        else:
            return False

    def decode_and_compute_shift(self, value: int, shift: int) -> tuple[int, bint]:
        immediate = not get_bit(shift, 0)
        if immediate:
            shift_amount = get_bits(shift, 3, 7)
        else:
            shift_reg = get_bits(shift, 4, 7)
            shift_amount = self.regs.get(shift_reg) & 255
        shift_type = get_bits(shift, 1, 2)
        return self.compute_shift(value, shift_type, shift_amount, immediate)

    def compute_shift(self, value: int, shift_type: ShiftType | int, shift_amount: int, immediate: bint) -> tuple[int, bint]:
        if not immediate and shift_amount == 0:
            return (value, self.regs.cpsr.get_carry_flag())
        result = value
        carry_out = False
        if shift_type == ShiftType.LSL:
            if shift_amount == 0:
                carry_out = self.regs.cpsr.get_carry_flag()
            elif shift_amount < 32:
                carry_out = get_bit(value, 32 - shift_amount)
                mask = 4294967295
                result = value << shift_amount & mask
            elif shift_amount == 32:
                carry_out = get_bit(value, 0)
                result = 0
            else:
                carry_out = False
                result = 0
        elif shift_type == ShiftType.LSR:
            if shift_amount == 0:
                carry_out = get_bit(value, 31)
                result = 0
            elif shift_amount < 32:
                carry_out = get_bit(value, shift_amount - 1)
                result = value >> shift_amount
            elif shift_amount == 32:
                carry_out = get_bit(value, 31)
                result = 0
            else:
                carry_out = False
                result = 0
        elif shift_type == ShiftType.ASR:
            if shift_amount == 0 or shift_amount >= 32:
                carry_out = get_bit(value, 31)
                result = 4294967295 if carry_out else 0
            else:
                carry_out = get_bit(value, shift_amount - 1)
                mask = 4294967295
                result = extend_sign_32(value) >> shift_amount & mask
        elif shift_type == ShiftType.ROR:
            if shift_amount > 32:
                shift_amount = (shift_amount - 1) % 32 + 1
            if shift_amount == 0:
                carry_out = get_bit(value, 0)
                result = value >> 1 | self.regs.cpsr.get_carry_flag() << 31
            else:
                carry_out = get_bit(value, shift_amount - 1)
                result = ror_32(value, get_bits(shift_amount, 0, 4))
        return (result, carry_out)


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.arm.constants import DataTransferOpcode
from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import get_bits, get_bit, add_32
# endif


def arm_halfword_data_transfer(cpu: CPU, instr: int):
    """Execute a Halfword Data Transfer instruction (LDRH, STRH, etc.)"""

    rn = get_bits(instr, 16, 19)  # Base reg
    rd = get_bits(instr, 12, 15)  # Source/dest reg

    pre_post_bit = get_bit(instr, 24)
    up_down_bit = get_bit(instr, 23)
    immediate_bit = get_bit(instr, 22)
    write_back_bit = get_bit(instr, 21)
    load_store_bit = get_bit(instr, 20)

    offset = (
        (get_bits(instr, 8, 11) << 4) | get_bits(instr, 0, 3)
        if immediate_bit
        else cpu.regs.get(get_bits(instr, 0, 3))
    )

    base = cpu.regs.get(rn)
    address = add_32(base, offset if up_down_bit else -offset)
    effective_address = address if pre_post_bit else base

    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL

    opcode = get_bits(instr, 5, 6)

    if load_store_bit:  # Load
        # When post-indexing, write back is always enabled
        # When pre-indexing, write back is optional (controlled by W bit)
        if not pre_post_bit or write_back_bit:
            cpu.regs.set(rn, address)

        if opcode == DataTransferOpcode.LDRH:
            cpu.regs.set(rd, cpu.memory.read_16_ror(effective_address, MemoryAccess.NON_SEQUENTIAL))
        elif opcode == DataTransferOpcode.LDRSB:
            cpu.regs.set(
                rd, cpu.memory.read_8_signed(effective_address, MemoryAccess.NON_SEQUENTIAL)
            )
        elif opcode == DataTransferOpcode.LDRSH:
            cpu.regs.set(
                rd, cpu.memory.read_16_signed(effective_address, MemoryAccess.NON_SEQUENTIAL)
            )

        cpu.scheduler.idle(1)

        if rd == cpu.regs.PC:
            cpu.flush_pipeline()

    else:  # Store
        if opcode == DataTransferOpcode.STRH:
            cpu.memory.write_16(effective_address, cpu.regs.get(rd), MemoryAccess.NON_SEQUENTIAL)
        elif opcode == DataTransferOpcode.LDRD:
            pass  # LDRD not implemented on ARM7
        elif opcode == DataTransferOpcode.STRD:
            pass  # STRD not implemented on ARM7

        # When post-indexing, write back is always enabled
        # When pre-indexing, write back is optional (controlled by W bit)
        if not pre_post_bit or write_back_bit:
            cpu.regs.set(rn, address)



# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.arm.constants import MultiplyLongOpcode
from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import get_bit, get_bits, sign_32, extend_sign_32
# endif


def arm_multiply(cpu: CPU, instr: int):
    """Execute a Multiply instruction"""

    rd = get_bits(instr, 16, 19)  # Destination reg
    rn = get_bits(instr, 12, 15)  # Accumulate reg
    rs = get_bits(instr, 8, 11)  # Operand reg
    rm = get_bits(instr, 0, 3)  # Operand reg

    accumulate = get_bit(instr, 21)
    set_cond_codes = get_bit(instr, 20)

    arm_multiply_idle(cpu, cpu.regs.get(rs), True)

    mask = 0xFFFFFFFF

    if accumulate:
        cpu.scheduler.idle(1)
        cpu.regs.set(rd, (cpu.regs.get(rm) * cpu.regs.get(rs) + cpu.regs.get(rn)) & mask)
    else:
        cpu.regs.set(rd, (cpu.regs.get(rm) * cpu.regs.get(rs)) & mask)

    if set_cond_codes:
        cpu.regs.cpsr.sign_flag = sign_32(cpu.regs.get(rd))
        cpu.regs.cpsr.zero_flag = cpu.regs.get(rd) == 0

    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL


def arm_multiply_long(cpu: CPU, instr: int):
    """Execute a Multiply Long instruction"""

    rd_hi = get_bits(instr, 16, 19)  # Upper bits of destination reg
    rd_lo = get_bits(instr, 12, 15)  # Lower bits of destination reg
    rs = get_bits(instr, 8, 11)  # Operand reg
    rm = get_bits(instr, 0, 3)  # Operand reg

    rd_hi_val = cpu.regs.get(rd_hi)
    rd_lo_val = cpu.regs.get(rd_lo)
    rs_val = cpu.regs.get(rs)
    rm_val = cpu.regs.get(rm)

    opcode = get_bits(instr, 21, 22)
    set_cond_codes = get_bit(instr, 20)

    cpu.scheduler.idle(1)

    if opcode == MultiplyLongOpcode.UMULL:
        arm_multiply_idle(cpu, rs_val, False)

        result = rm_val * rs_val

    elif opcode == MultiplyLongOpcode.UMLAL:
        arm_multiply_idle(cpu, rs_val, False)
        cpu.scheduler.idle(1)

        result = (rd_hi_val << 32) | rd_lo_val
        result += rm_val * rs_val

    elif opcode == MultiplyLongOpcode.SMULL:
        arm_multiply_idle(cpu, rs_val, True)

        result = extend_sign_32(rm_val) * extend_sign_32(rs_val)

    elif opcode == MultiplyLongOpcode.SMLAL:
        arm_multiply_idle(cpu, rs_val, True)
        cpu.scheduler.idle(1)

        result = (rd_hi_val << 32) | rd_lo_val
        result += extend_sign_32(rm_val) * extend_sign_32(rs_val)

    # ifndef CYTHON
    else:
        assert False, "Unreachable, opcode must be one of MultiplyLongOpcode"
    # endif

    mask = 0xFFFFFFFF
    cpu.regs.set(rd_lo, result & mask)
    cpu.regs.set(rd_hi, (result >> 32) & mask)

    if set_cond_codes:
        cpu.regs.cpsr.sign_flag = sign_32(cpu.regs.get(rd_hi))
        cpu.regs.cpsr.zero_flag = cpu.regs.get(rd_hi) == 0 and cpu.regs.get(rd_lo) == 0

    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL


def arm_multiply_idle(cpu: CPU, rs: int, signed: bool):
    """
    Idle for the number of cycles it takes to execute a multiply instruction.
    The number of cycles depends on how many MSBs of the Rs operand are
    "all 0" (if unsigned) or "all 0 or all 1" (if signed).
    """

    cycles = 1
    mask = 0xFFFFFF00
    while mask != 0:
        rs &= mask
        if rs == 0 or (signed and rs == mask):
            break
        cycles += 1
        mask <<= 8

    cpu.scheduler.idle(cycles)



from ctypes import c_void_p

class PPU:
    """
    Picture Processing Unit.
    """

    def __init__(self, scheduler: Scheduler, interrupt_controller: InterruptController):
        self.scheduler = scheduler
        self.interrupt_controller = interrupt_controller
        self.rendering_enabled = True
        self.display_control = DisplayControlRegister()
        self.display_status = DisplayStatusRegister()
        self.vcount = 0
        self.bg_control_0 = BackgroundControlRegister()
        self.bg_control_1 = BackgroundControlRegister()
        self.bg_control_2 = BackgroundControlRegister()
        self.bg_control_3 = BackgroundControlRegister()
        self.bg_offset_h = array('I', [0] * NUM_BACKGROUNDS)
        self.bg_offset_v = array('I', [0] * NUM_BACKGROUNDS)
        self.window_v_min = array('I', [0] * NUM_PRIMARY_WINDOWS)
        self.window_v_max = array('I', [0] * NUM_PRIMARY_WINDOWS)
        self.window_h_min = array('I', [0] * NUM_PRIMARY_WINDOWS)
        self.window_h_max = array('I', [0] * NUM_PRIMARY_WINDOWS)
        self.window_control_0 = WindowControlRegister()
        self.window_control_1 = WindowControlRegister()
        self.window_control_obj = WindowControlRegister()
        self.window_control_out = WindowControlRegister()
        self.window_mask_0 = array('H', [0] * DISPLAY_WIDTH)
        self.window_mask_1 = array('H', [0] * DISPLAY_WIDTH)
        self.window_mask_obj = array('H', [0] * DISPLAY_WIDTH)
        self.blend_control = BlendControlRegister()
        self.blend_alpha = BlendAlphaRegister()
        self.blend_brightness = BlendBrightnessRegister()
        self.blend_mask_obj = [[False] * DISPLAY_WIDTH for _ in range(NUM_PRIORITIES)]
        self.memory = VideoMemory(self.display_control)
        self.bg_scanline = [array('H', [0] * DISPLAY_WIDTH) for _ in range(NUM_BACKGROUNDS)]
        self.obj_scanline = [array('H', [0] * DISPLAY_WIDTH) for _ in range(NUM_PRIORITIES)]
        self.scanline = array('H', [0] * DISPLAY_WIDTH)
        self.scanline_top_colour = array('H', [0] * DISPLAY_WIDTH)
        self.scanline_top_layer = array('I', [LayerType.BACKDROP] * DISPLAY_WIDTH)
        self.front_buffer = array('H', [0] * (DISPLAY_WIDTH * DISPLAY_HEIGHT))
        self.back_buffer = array('H', [0] * (DISPLAY_WIDTH * DISPLAY_HEIGHT))
        self.front_buffer_ptr = c_void_p(self.front_buffer.buffer_info()[0])
        self.back_buffer_ptr = c_void_p(self.back_buffer.buffer_info()[0])
        self.scheduler.schedule(self.hblank_start, CYCLES_HDRAW, EventTrigger.IMMEDIATELY)

    def get_frame_buffer_ptr(self) -> c_void_p:
        return self.front_buffer_ptr

    def hblank_start(self):
        self.display_status.set_hblank_status(True)
        if self.vcount < DISPLAY_HEIGHT:
            if self.rendering_enabled:
                self._init_layers()
                self._calc_window_masks()
                if not self.display_control.get_force_blank():
                    self._render_backgrounds()
                    self._render_objects()
                self._merge_layers()
            self.scheduler.trigger(EventTrigger.EventTrigger_HBLANK)
        if self.display_status.get_hblank_irq():
            self.interrupt_controller.signal(Interrupt.HBLANK)
        self.scheduler.schedule(self.hblank_end, CYCLES_HBLANK, EventTrigger.IMMEDIATELY)

    def hblank_end(self):
        self.vcount += 1
        if self.vcount >= DISPLAY_FULL_HEIGHT:
            self.vcount = 0
        self.display_status.set_hblank_status(False)
        self.display_status.set_vblank_status(self.vcount >= DISPLAY_HEIGHT)
        self.display_status.set_vcount_trigger_status(self.vcount == self.display_status.get_vcount_trigger_value())
        if self.vcount == DISPLAY_HEIGHT:
            self.front_buffer, self.back_buffer = (self.back_buffer, self.front_buffer)
            self.front_buffer_ptr, self.back_buffer_ptr = (self.back_buffer_ptr, self.front_buffer_ptr)
            if self.display_status.get_vblank_irq():
                self.interrupt_controller.signal(Interrupt.VBLANK)
            self.scheduler.trigger(EventTrigger.EventTrigger_VBLANK)
        if self.display_status.get_vcount_irq() and self.display_status.get_vcount_trigger_status():
            self.interrupt_controller.signal(Interrupt.VCOUNT)
        self.scheduler.schedule(self.hblank_start, CYCLES_HDRAW, EventTrigger.IMMEDIATELY)

    def _init_layers(self):
        backdrop_colour = 0 if self.display_control.get_force_blank() else self.memory.read_16_palram(0)
        for x in range(DISPLAY_WIDTH):
            self.scanline[x] = backdrop_colour
            self.scanline_top_colour[x] = backdrop_colour
            self.scanline_top_layer[x] = LayerType.BACKDROP
        for scanline in self.bg_scanline:
            for x in range(DISPLAY_WIDTH):
                scanline[x] = TRANSPARENT_COLOUR
        for scanline in self.obj_scanline:
            for x in range(DISPLAY_WIDTH):
                scanline[x] = TRANSPARENT_COLOUR
        for blend_mask in self.blend_mask_obj:
            for x in range(DISPLAY_WIDTH):
                blend_mask[x] = False

    def _render_backgrounds(self):
        video_mode = self.display_control.get_video_mode()
        if video_mode == VideoMode.MODE_0:
            for bg_num in range(NUM_BACKGROUNDS):
                self._render_background_text(bg_num)
        elif video_mode == VideoMode.MODE_1:
            self._render_background_affine(2)
            self._render_background_text(1)
            self._render_background_text(0)
        elif video_mode == VideoMode.MODE_2:
            self._render_background_affine(3)
            self._render_background_affine(2)
        elif video_mode == VideoMode.MODE_3:
            self._render_background_bitmap(False, False)
        elif video_mode == VideoMode.MODE_4:
            self._render_background_bitmap(True, False)
        elif video_mode == VideoMode.MODE_5:
            self._render_background_bitmap(False, True)

    def _render_background_text(self, bg_num: int):
        if not self.display_control.display_bg(bg_num):
            return
        bg_control = self._get_bg_control(bg_num)
        rel_y = self.vcount + self.bg_offset_v[bg_num]
        tile_y = rel_y // TILE_HEIGHT
        pixel_y = rel_y % TILE_HEIGHT
        for x in range(DISPLAY_WIDTH):
            rel_x = x + self.bg_offset_h[bg_num]
            tile_x = rel_x // TILE_WIDTH
            pixel_x = rel_x % TILE_WIDTH
            tile_map_block = bg_control.get_tile_map_block()
            if bg_control.get_size() == 1:
                if tile_x & TILE_MAP_BLOCK_WIDTH:
                    tile_map_block += 1
            elif bg_control.get_size() == 2:
                if tile_y & TILE_MAP_BLOCK_HEIGHT:
                    tile_map_block += 1
            elif bg_control.get_size() == 3:
                if tile_x & TILE_MAP_BLOCK_WIDTH:
                    tile_map_block += 1
                if tile_y & TILE_MAP_BLOCK_HEIGHT:
                    tile_map_block += 2
            tile_map_entry_address = tile_y % TILE_MAP_BLOCK_HEIGHT * TILE_MAP_BLOCK_WIDTH
            tile_map_entry_address += tile_x % TILE_MAP_BLOCK_WIDTH
            tile_map_entry_address *= TILE_MAP_ENTRY_SIZE
            tile_map_entry_address += tile_map_block * TILE_MAP_BLOCK_SIZE
            tile_map_entry = self.memory.read_16_vram(tile_map_entry_address)
            tile_num = get_bits(tile_map_entry, 0, 9)
            tile_flip_h = get_bit(tile_map_entry, 10)
            tile_flip_v = get_bit(tile_map_entry, 11)
            palette_num = get_bits(tile_map_entry, 12, 15)
            pixel_x_flipped = TILE_WIDTH - pixel_x - 1 if tile_flip_h else pixel_x
            pixel_y_flipped = TILE_HEIGHT - pixel_y - 1 if tile_flip_v else pixel_y
            palette_index = self._get_palette_index(bg_control.get_tile_set_block() * TILE_SET_BLOCK_SIZE, tile_num, pixel_x_flipped, pixel_y_flipped, bg_control.get_colour_256(), palette_num)
            if palette_index:
                colour = self.memory.read_16_palram(palette_index * COLOUR_SIZE)
                self.bg_scanline[bg_num][x] = colour

    def _render_background_affine(self, bg_num: int):
        if not self.display_control.display_bg(bg_num):
            return

    def _render_background_bitmap(self, paletted, small):
        if not self.display_control.display_bg(2):
            return
        page_offset = VRAM_PAGE_OFFSET if self.display_control.get_page_select() else 0
        display_width = SMALL_DISPLAY_WIDTH if small else DISPLAY_WIDTH
        display_height = SMALL_DISPLAY_HEIGHT if small else DISPLAY_HEIGHT
        if paletted:
            for x in range(display_width):
                palette_index = self.memory.read_8_vram(page_offset + display_width * self.vcount + x)
                if palette_index:
                    colour = self.memory.read_16_palram(palette_index * COLOUR_SIZE)
                    self.bg_scanline[2][x] = colour
        else:
            for x in range(display_width):
                vram_index = display_width * self.vcount + x
                colour = self.memory.read_16_vram(vram_index * COLOUR_SIZE)
                self.bg_scanline[2][x] = colour

    def _render_objects(self):
        if not self.display_control.get_display_obj():
            return
        for obj_index in range(MAX_NUM_OBJECTS - 1, -1, -1):
            obj_address = obj_index * OAM_ENTRY_SIZE
            self._render_object(Object(self.memory.read_16_oam(obj_address), self.memory.read_16_oam(obj_address + 2), self.memory.read_16_oam(obj_address + 4)))

    def _render_object(self, obj: Object):
        if obj.get_disabled():
            return
        video_mode = self.display_control.get_video_mode()
        bitmapped = video_mode in [VideoMode.MODE_3, VideoMode.MODE_4, VideoMode.MODE_5]
        if bitmapped and obj.get_tile_index() < 512:
            return
        obj_x = extend_sign_9(obj.get_x())
        obj_y = obj.get_y()
        obj_w, obj_h = obj.get_size()
        obj_flip_h, obj_flip_v = (obj.get_flip_horizontal(), obj.get_flip_vertical())
        if obj_y + obj_h > 256:
            obj_y -= 256
        if self.vcount < obj_y or self.vcount >= obj_y + obj_h:
            return
        tile_row_len = obj_w // TILE_WIDTH if self.display_control.get_obj_vram_dimension() else 32
        offset_y = self.vcount - obj_y
        rel_y = obj_y - offset_y - 1 if obj_flip_v else offset_y
        for offset_x in range(0, obj_w):
            win_x = add_32(obj_x, offset_x)
            if sign_32(win_x) or win_x >= DISPLAY_WIDTH:
                continue
            rel_x = obj_w - offset_x - 1 if obj_flip_h else offset_x
            tile_x = rel_x // TILE_WIDTH
            tile_y = rel_y // TILE_HEIGHT
            tile_index = obj.get_tile_index() + tile_y * tile_row_len + tile_x
            pixel_x = rel_x % TILE_WIDTH
            pixel_y = rel_y % TILE_HEIGHT
            palette_index = self._get_palette_index(OBJ_TILE_SET_OFFSET, tile_index, pixel_x, pixel_y, obj.get_colour_256(), obj.get_palette_num())
            if palette_index:
                if obj.get_mode() == ObjectMode.WINDOW:
                    self.window_mask_obj[win_x] = 1
                else:
                    colour = self.memory.read_16_palram(OBJ_PALETTE_OFFSET + palette_index * COLOUR_SIZE)
                    self.obj_scanline[obj.get_priority()][win_x] = colour
                    self.blend_mask_obj[obj.get_priority()][win_x] = obj.get_mode() == ObjectMode.BLEND

    def _merge_layers(self):
        for priority in range(NUM_PRIORITIES - 1, -1, -1):
            for bg_num in range(NUM_BACKGROUNDS - 1, -1, -1):
                if self.display_control.display_bg(bg_num) and self._get_bg_control(bg_num).get_priority() == priority:
                    self._merge_layer(self.bg_scanline[bg_num], bg_num, priority)
            if self.display_control.get_display_obj():
                self._merge_layer(self.obj_scanline[priority], LayerType.OBJ, priority)
        for x in range(DISPLAY_WIDTH):
            self.back_buffer[DISPLAY_WIDTH * self.vcount + x] = self.scanline[x]

    def _merge_layer(self, layer_scanline, layer_type: LayerType | int, priority: int):
        windowing_enabled = self.display_control.display_window(WindowIndex.WIN_0) or self.display_control.display_window(WindowIndex.WIN_1) or self.display_control.display_window(WindowIndex.WIN_OBJ)
        blend_mode = self.blend_control.get_blend_mode() if self.blend_control.blend_source(layer_type) else BlendMode.OFF
        coefficient_source = self.blend_alpha.get_coefficient_source()
        coefficient_target = self.blend_alpha.get_coefficient_target()
        brightness = self.blend_brightness.get_brightness()
        for x in range(DISPLAY_WIDTH):
            pixel_blend_mode = blend_mode
            if windowing_enabled:
                window_control = self._get_window_for_pixel(x)
                if not window_control.display_layer(layer_type):
                    continue
                if not window_control.get_enable_blending():
                    pixel_blend_mode = BlendMode.OFF
            if layer_type == LayerType.OBJ and self.blend_mask_obj[priority][x]:
                pixel_blend_mode = BlendMode.ALPHA
            colour = layer_scanline[x]
            if colour == TRANSPARENT_COLOUR:
                continue
            if pixel_blend_mode == BlendMode.OFF:
                self.scanline[x] = colour
            elif pixel_blend_mode == BlendMode.ALPHA:
                if self.blend_control.blend_target(self.scanline_top_layer[x]):
                    self.scanline[x] = blend_colours(colour, self.scanline_top_colour[x], coefficient_source, coefficient_target)
                else:
                    self.scanline[x] = colour
            elif pixel_blend_mode == BlendMode.LIGHTEN:
                self.scanline[x] = colour
            elif pixel_blend_mode == BlendMode.DARKEN:
                self.scanline[x] = colour
            self.scanline_top_colour[x] = colour
            self.scanline_top_layer[x] = layer_type

    def _calc_window_masks(self):
        for window in (WindowIndex.WIN_0, WindowIndex.WIN_1):
            window_mask = self.window_mask_0 if window == WindowIndex.WIN_0 else self.window_mask_1
            h_min = self.window_h_min[int(window)]
            h_max = self.window_h_max[int(window)]
            v_min = self.window_v_min[int(window)]
            v_max = self.window_v_max[int(window)]
            inside_v_bounds = v_min <= self.vcount < v_max if v_min <= v_max else self.vcount < v_max or self.vcount >= v_min
            if not self.display_control.display_window(window) or not inside_v_bounds:
                for x in range(DISPLAY_WIDTH):
                    window_mask[x] = 0
            elif h_min <= h_max:
                for x in range(DISPLAY_WIDTH):
                    window_mask[x] = h_min <= x < h_max
            else:
                for x in range(DISPLAY_WIDTH):
                    window_mask[x] = x < h_max or x >= h_min
        for x in range(DISPLAY_WIDTH):
            self.window_mask_obj[x] = 0

    def _get_window_for_pixel(self, x):
        if self.window_mask_0[x]:
            return self.window_control_0
        elif self.window_mask_1[x]:
            return self.window_control_1
        elif self.window_mask_obj[x]:
            return self.window_control_obj
        else:
            return self.window_control_out

    def _get_palette_index(self, tile_base_address: int, tile_index: int, pixel_x: int, pixel_y: int, colour_256: bint, palette_num: int) -> int:
        if colour_256:
            pixel_address = tile_base_address
            pixel_address += tile_index * TILE_SIZE + pixel_y * TILE_WIDTH + pixel_x
            palette_index = self.memory.read_8_vram(pixel_address)
        else:
            pixel_address = tile_base_address
            pixel_address += (tile_index * TILE_SIZE + pixel_y * TILE_WIDTH + pixel_x) // 2
            palette_index = self.memory.read_8_vram(pixel_address) >> pixel_x % 2 * 4 & 15
            if palette_index == 0:
                return 0
            palette_index += palette_num * 16
        return palette_index

    def _get_bg_control(self, bg_num: int) -> BackgroundControlRegister:
        if bg_num == 0:
            return self.bg_control_0
        elif bg_num == 1:
            return self.bg_control_1
        elif bg_num == 2:
            return self.bg_control_2
        elif bg_num == 3:
            return self.bg_control_3

class Object:
    """
    Corresponds to an OAM entry representing an object (moveable sprite).
    Each entry consists of three 16-bit attributes.
    """

    def __init__(self, attr_0: int, attr_1: int, attr_2: int):
        self.SIZES = [[(8, 8), (16, 16), (32, 32), (64, 64)], [(16, 8), (32, 8), (32, 16), (64, 32)], [(8, 16), (8, 32), (16, 32), (32, 64)]]
        self.attr_0 = attr_0
        self.attr_1 = attr_1
        self.attr_2 = attr_2

    def get_x(self) -> int:
        return get_bits(self.attr_1, 0, 8)

    def get_y(self) -> int:
        return get_bits(self.attr_0, 0, 7)

    def get_affine(self) -> bint:
        return get_bit(self.attr_0, 8)

    def get_double_size(self) -> bint:
        return get_bit(self.attr_0, 9)

    def get_disabled(self) -> bint:
        return get_bit(self.attr_0, 9) & ~self.get_affine()

    def get_mode(self) -> ObjectMode | int:
        return get_bits(self.attr_0, 10, 11)

    def get_mosaic(self) -> bint:
        return get_bit(self.attr_0, 12)

    def get_colour_256(self) -> bint:
        return get_bit(self.attr_0, 13)

    def get_shape(self) -> ObjectShape | int:
        return get_bits(self.attr_0, 14, 15)

    def get_flip_horizontal(self) -> bint:
        return get_bit(self.attr_1, 12) & ~self.get_affine()

    def get_flip_vertical(self) -> bint:
        return get_bit(self.attr_1, 13) & ~self.get_affine()

    def get_size(self) -> tuple[int, int]:
        size = get_bits(self.attr_1, 14, 15)
        return self.SIZES[self.get_shape()][size]

    def get_tile_index(self) -> int:
        return get_bits(self.attr_2, 0, 9)

    def get_priority(self) -> int:
        return get_bits(self.attr_2, 10, 11)

    def get_palette_num(self) -> int:
        return get_bits(self.attr_2, 12, 15)

def get_rgb_channels(colour: int) -> tuple[int, int, int]:
    r = colour & 31
    g = colour >> 5 & 31
    b = colour >> 10 & 31
    return (r, g, b)

def blend_colours(colour_a: int, colour_b: int, coefficient_a: int, coefficient_b: int) -> int:
    r_a, g_a, b_a = get_rgb_channels(colour_a)
    r_b, g_b, b_b = get_rgb_channels(colour_b)
    r = min(31, r_a * coefficient_a + r_b * coefficient_b >> 4)
    g = min(31, g_a * coefficient_a + g_b * coefficient_b >> 4)
    b = min(31, b_a * coefficient_a + b_b * coefficient_b >> 4)
    return r | g << 5 | b << 10


# ifndef CYTHON
from pyboy_advance.app.constants import WindowEvent
from pyboy_advance.constants import Interrupt, Key
from pyboy_advance.interrupt_controller import InterruptController
from pyboy_advance.utils import bint, get_bit, get_bits
# endif


class Keypad:
    def __init__(self, interrupt_controller: InterruptController):
        self.interrupt_controller = interrupt_controller

        # Initialize all keys to "released" state (inputs are active low)
        self.key_input = 0b1111111111

        self.key_control = KeypadInterruptControlRegister()

    def press_key(self, key: Key):
        self.key_input &= ~key

        if self.key_control.irq_enable and self._evaluate_irq_condition():
            self.interrupt_controller.signal(Interrupt.KEYPAD)

    def release_key(self, key: Key):
        self.key_input |= key

        if self.key_control.irq_enable and self._evaluate_irq_condition():
            self.interrupt_controller.signal(Interrupt.KEYPAD)

    def process_window_event(self, event: WindowEvent):
        if event == WindowEvent.PRESS_BUTTON_A:
            self.press_key(Key.BUTTON_A)
        elif event == WindowEvent.PRESS_BUTTON_B:
            self.press_key(Key.BUTTON_B)
        elif event == WindowEvent.PRESS_BUTTON_SELECT:
            self.press_key(Key.BUTTON_SELECT)
        elif event == WindowEvent.PRESS_BUTTON_START:
            self.press_key(Key.BUTTON_START)
        elif event == WindowEvent.PRESS_DPAD_RIGHT:
            self.press_key(Key.DPAD_RIGHT)
        elif event == WindowEvent.PRESS_DPAD_LEFT:
            self.press_key(Key.DPAD_LEFT)
        elif event == WindowEvent.PRESS_DPAD_UP:
            self.press_key(Key.DPAD_UP)
        elif event == WindowEvent.PRESS_DPAD_DOWN:
            self.press_key(Key.DPAD_DOWN)
        elif event == WindowEvent.PRESS_SHOULDER_RIGHT:
            self.press_key(Key.SHOULDER_RIGHT)
        elif event == WindowEvent.PRESS_SHOULDER_LEFT:
            self.press_key(Key.SHOULDER_LEFT)
        elif event == WindowEvent.RELEASE_BUTTON_A:
            self.release_key(Key.BUTTON_A)
        elif event == WindowEvent.RELEASE_BUTTON_B:
            self.release_key(Key.BUTTON_B)
        elif event == WindowEvent.RELEASE_BUTTON_SELECT:
            self.release_key(Key.BUTTON_SELECT)
        elif event == WindowEvent.RELEASE_BUTTON_START:
            self.release_key(Key.BUTTON_START)
        elif event == WindowEvent.RELEASE_DPAD_RIGHT:
            self.release_key(Key.DPAD_RIGHT)
        elif event == WindowEvent.RELEASE_DPAD_LEFT:
            self.release_key(Key.DPAD_LEFT)
        elif event == WindowEvent.RELEASE_DPAD_UP:
            self.release_key(Key.DPAD_UP)
        elif event == WindowEvent.RELEASE_DPAD_DOWN:
            self.release_key(Key.DPAD_DOWN)
        elif event == WindowEvent.RELEASE_SHOULDER_RIGHT:
            self.release_key(Key.SHOULDER_RIGHT)
        elif event == WindowEvent.RELEASE_SHOULDER_LEFT:
            self.release_key(Key.SHOULDER_LEFT)
        # ifndef CYTHON
        else:
            raise ValueError(f"WindowEvent {event.name} is not a key event")
        # endif

    def _evaluate_irq_condition(self) -> bool:
        if self.key_control.irq_if_all:
            # Logical AND mode; interrupt requested if ALL of selected keys are pressed
            return (~self.key_input & self.key_control.key_select) == self.key_control.key_select
        else:
            # Logical OR mode; interrupt requested if ANY of selected keys are pressed
            return (~self.key_input & self.key_control.key_select) != 0


class KeypadInterruptControlRegister:
    def __init__(self):
        self.reg = 0

    @property
    def key_select(self) -> int:
        return get_bits(self.reg, 0, 9)

    @property
    def irq_enable(self) -> bint:
        return get_bit(self.reg, 14)

    @property
    def irq_if_all(self) -> bint:
        return get_bit(self.reg, 15)



class Registers:
    """
    ARM7TDMI register set. All registers are 32 bits wide.

    Only 16 registers are visible at any given time.
    There are 20 "banked" registers that get swapped in whenever the CPU
    switches to various privileged modes.
    """

    def __init__(self):
        self.SP = 13
        self.LR = 14
        self.PC = 15
        self.BANKED_GPR_RANGE_START = 8
        self.BANKED_GPR_RANGE_END = 12
        self.BANKED_GPR_RANGE_LEN = self.BANKED_GPR_RANGE_END - self.BANKED_GPR_RANGE_START + 1
        self.regs = array('I', [0] * 16)
        self.cpsr = ProgramStatusRegister()
        self.spsr = ProgramStatusRegister()
        self.banked_old_gpr = array('I', [0] * self.BANKED_GPR_RANGE_LEN)
        self.banked_fiq_gpr = array('I', [0] * self.BANKED_GPR_RANGE_LEN)
        self.banked_sp = array('I', [0] * 6)
        self.banked_lr = array('I', [0] * 6)
        self.banked_spsr_user = ProgramStatusRegister()
        self.banked_spsr_fiq = ProgramStatusRegister()
        self.banked_spsr_irq = ProgramStatusRegister()
        self.banked_spsr_swi = ProgramStatusRegister()
        self.banked_spsr_abort = ProgramStatusRegister()
        self.banked_spsr_undefined = ProgramStatusRegister()

    def get(self, reg: int):
        return self.regs[reg]

    def set(self, reg: int, value: int):
        self.regs[reg] = value

    def get_sp(self):
        return self.regs[self.SP]

    def set_sp(self, value: int):
        self.regs[self.SP] = value

    def get_lr(self):
        return self.regs[self.LR]

    def set_lr(self, value: int):
        self.regs[self.LR] = value

    def get_pc(self):
        return self.regs[self.PC]

    def set_pc(self, value: int):
        self.regs[self.PC] = value

    def switch_mode(self, new_mode: CPUMode):
        """
        Switch from the current mode to the specified mode.

        Saves the current register values to the current mode's bank and
        replaces them with values from the new mode's bank.
        """
        if new_mode == self.cpsr.get_mode():
            return
        old_mode = self.cpsr.get_mode()
        self.cpsr.set_mode(new_mode)
        old_bank_index = self.get_bank_index(old_mode)
        new_bank_index = self.get_bank_index(new_mode)
        self.banked_sp[old_bank_index] = self.get_sp()
        self.banked_lr[old_bank_index] = self.get_lr()
        self.get_banked_spsr(old_bank_index).reg = self.spsr.reg
        self.set_sp(self.banked_sp[new_bank_index])
        self.set_lr(self.banked_lr[new_bank_index])
        self.spsr.reg = self.get_banked_spsr(new_bank_index).reg
        if new_mode == CPUMode.FIQ:
            for i in range(self.BANKED_GPR_RANGE_LEN):
                self.banked_old_gpr[i] = self.regs[i + self.BANKED_GPR_RANGE_START]
                self.regs[i + self.BANKED_GPR_RANGE_START] = self.banked_fiq_gpr[i]
        elif old_mode == CPUMode.FIQ:
            for i in range(self.BANKED_GPR_RANGE_LEN):
                self.banked_fiq_gpr[i] = self.regs[i + self.BANKED_GPR_RANGE_START]
                self.regs[i + self.BANKED_GPR_RANGE_START] = self.banked_old_gpr[i]

    def get_bank_index(self, mode: CPUMode) -> BankIndex:
        if mode == CPUMode.SYSTEM or mode == CPUMode.USER:
            return BankIndex.BANK_SYSTEM_USER
        elif mode == CPUMode.FIQ:
            return BankIndex.BANK_FIQ
        elif mode == CPUMode.IRQ:
            return BankIndex.BANK_IRQ
        elif mode == CPUMode.SWI:
            return BankIndex.BANK_SWI
        elif mode == CPUMode.ABORT:
            return BankIndex.BANK_ABORT
        elif mode == CPUMode.UNDEFINED:
            return BankIndex.BANK_UNDEFINED

    def get_banked_spsr(self, bank_index: BankIndex) -> ProgramStatusRegister:
        if bank_index == BankIndex.BANK_SYSTEM_USER:
            return self.banked_spsr_user
        if bank_index == BankIndex.BANK_FIQ:
            return self.banked_spsr_fiq
        elif bank_index == BankIndex.BANK_IRQ:
            return self.banked_spsr_irq
        elif bank_index == BankIndex.BANK_SWI:
            return self.banked_spsr_swi
        elif bank_index == BankIndex.BANK_ABORT:
            return self.banked_spsr_abort
        elif bank_index == BankIndex.BANK_UNDEFINED:
            return self.banked_spsr_undefined

class ProgramStatusRegister:

    def __init__(self, value: int=0):
        self.reg = value

    def get_mode(self) -> CPUMode | int:
        return self.reg & 31

    def set_mode(self, mode: CPUMode):
        self.reg = self.reg & ~31 | mode

    def get_state(self) -> CPUState | int:
        return get_bit(self.reg, 5)

    def set_state(self, state: CPUState):
        self.reg = set_bit(self.reg, 5, state)

    def get_sign_flag(self) -> bint:
        return get_bit(self.reg, 31)

    def set_sign_flag(self, value: bint):
        self.reg = set_bit(self.reg, 31, value)

    def get_zero_flag(self) -> bint:
        return get_bit(self.reg, 30)

    def set_zero_flag(self, value: bint):
        self.reg = set_bit(self.reg, 30, value)

    def get_carry_flag(self) -> bint:
        return get_bit(self.reg, 29)

    def set_carry_flag(self, value: bint):
        self.reg = set_bit(self.reg, 29, value)

    def get_overflow_flag(self) -> bint:
        return get_bit(self.reg, 28)

    def set_overflow_flag(self, value: bint):
        self.reg = set_bit(self.reg, 28, value)

    def get_sticky_overflow_flag(self) -> bint:
        return get_bit(self.reg, 27)

    def set_sticky_overflow_flag(self, value: bint):
        self.reg = set_bit(self.reg, 27, value)

    def get_irq_disable(self) -> bint:
        return get_bit(self.reg, 7)

    def set_irq_disable(self, value: bint):
        self.reg = set_bit(self.reg, 7, value)

    def get_fiq_disable(self) -> bint:
        return get_bit(self.reg, 6)

    def set_fiq_disable(self, value: bint):
        self.reg = set_bit(self.reg, 6, value)


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.arm.alu import (
    arm_alu_sub,
    arm_alu_add,
    arm_alu_mov,
    arm_alu_adc,
    arm_alu_sbc,
    arm_alu_tst,
    arm_alu_cmp,
    arm_alu_cmn,
    arm_alu_orr,
    arm_alu_bic,
    arm_alu_mvn,
    arm_alu_and,
    arm_alu_eor,
)
from pyboy_advance.cpu.arm.mul import arm_multiply_idle
from pyboy_advance.cpu.constants import ShiftType
from pyboy_advance.cpu.thumb.constants import ThumbALUOpcode
from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import (
    get_bits,
    get_bit,
    sign_32,
    add_32,
)
# endif


def thumb_move_shifted_register(cpu: CPU, instr: int):
    """Execute a THUMB.1 instruction (move shifted register)"""

    rs = get_bits(instr, 3, 5)  # Source reg
    rd = get_bits(instr, 0, 2)  # Destination reg

    opcode = get_bits(instr, 11, 12)
    offset = get_bits(instr, 6, 10)

    result, carry = cpu.compute_shift(cpu.regs.get(rs), opcode, offset, True)

    cpu.regs.set(rd, result)
    cpu.regs.cpsr.sign_flag = sign_32(result)
    cpu.regs.cpsr.zero_flag = result == 0
    cpu.regs.cpsr.carry_flag = carry

    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def thumb_add_subtract(cpu: CPU, instr: int):
    """Execute a THUMB.2 instruction (add/subtract)"""

    rs = get_bits(instr, 3, 5)  # Source reg
    rd = get_bits(instr, 0, 2)  # Destination reg

    opcode = get_bits(instr, 9, 10)
    immediate_bit = get_bit(instr, 10)

    op1 = cpu.regs.get(rs)
    op2 = get_bits(instr, 6, 8) if immediate_bit else cpu.regs.get(get_bits(instr, 6, 8))

    if opcode == 0:  # ADD (Register)
        arm_alu_add(cpu, op1, op2, rd, True)
    elif opcode == 1:  # SUB (Register)
        arm_alu_sub(cpu, op1, op2, rd, True)
    elif opcode == 2:  # ADD (Immediate)
        arm_alu_add(cpu, op1, op2, rd, True)
    elif opcode == 3:  # SUB (Immediate)
        arm_alu_sub(cpu, op1, op2, rd, True)

    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def thumb_move_compare_add_subtract(cpu: CPU, instr: int):
    """Execute a THUMB.3 instruction (move/compare/add/subtract immediate)"""

    rd = get_bits(instr, 8, 10)  # Destination reg

    opcode = get_bits(instr, 11, 12)
    value = get_bits(instr, 0, 7)

    if opcode == 0:  # MOV
        arm_alu_mov(cpu, value, rd, True, cpu.regs.cpsr.carry_flag)
    elif opcode == 1:  # CMP
        arm_alu_cmp(cpu, cpu.regs.get(rd), value)
    elif opcode == 2:  # ADD
        arm_alu_add(cpu, cpu.regs.get(rd), value, rd, True)
    elif opcode == 3:  # SUB
        arm_alu_sub(cpu, cpu.regs.get(rd), value, rd, True)

    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def thumb_alu(cpu: CPU, instr: int):
    """Execute a THUMB.4 instruction (miscellaneous ALU operations)"""

    rs = get_bits(instr, 3, 5)
    rd = get_bits(instr, 0, 2)

    op1 = cpu.regs.get(rd)
    op2 = cpu.regs.get(rs)

    opcode = get_bits(instr, 6, 9)
    if opcode == ThumbALUOpcode.AND:
        arm_alu_and(cpu, op1, op2, rd, True, cpu.regs.cpsr.carry_flag)
    elif opcode == ThumbALUOpcode.EOR:
        arm_alu_eor(cpu, op1, op2, rd, True, cpu.regs.cpsr.carry_flag)
    elif opcode == ThumbALUOpcode.LSL:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.LSL)
    elif opcode == ThumbALUOpcode.LSR:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.LSR)
    elif opcode == ThumbALUOpcode.ASR:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.ASR)
    elif opcode == ThumbALUOpcode.ADC:
        arm_alu_adc(cpu, op1, op2, rd, True)
    elif opcode == ThumbALUOpcode.SBC:
        arm_alu_sbc(cpu, op1, op2, rd, True)
    elif opcode == ThumbALUOpcode.ROR:
        thumb_alu_shift(cpu, op1, op2, rd, ShiftType.ROR)
    elif opcode == ThumbALUOpcode.TST:
        arm_alu_tst(cpu, op1, op2, cpu.regs.cpsr.carry_flag)
    elif opcode == ThumbALUOpcode.NEG:
        arm_alu_sub(cpu, 0, op2, rd, True)
    elif opcode == ThumbALUOpcode.CMP:
        arm_alu_cmp(cpu, op1, op2)
    elif opcode == ThumbALUOpcode.CMN:
        arm_alu_cmn(cpu, op1, op2)
    elif opcode == ThumbALUOpcode.ORR:
        arm_alu_orr(cpu, op1, op2, rd, True, cpu.regs.cpsr.carry_flag)
    elif opcode == ThumbALUOpcode.MUL:
        thumb_alu_multiply(cpu, op1, op2, rd)
    elif opcode == ThumbALUOpcode.BIC:
        arm_alu_bic(cpu, op1, op2, rd, True, cpu.regs.cpsr.carry_flag)
    elif opcode == ThumbALUOpcode.MVN:
        arm_alu_mvn(cpu, op2, rd, True, cpu.regs.cpsr.carry_flag)

    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def thumb_alu_shift(cpu: CPU, value: int, shift: int, rd: int, shift_type: ShiftType):
    shift &= 0xFF
    result, carry = cpu.compute_shift(value, shift_type, shift, False)

    cpu.regs.set(rd, result)
    cpu.regs.cpsr.sign_flag = sign_32(result)
    cpu.regs.cpsr.zero_flag = result == 0
    cpu.regs.cpsr.carry_flag = carry

    cpu.scheduler.idle(1)


def thumb_alu_multiply(cpu: CPU, op1: int, op2: int, rd: int):
    mask = 0xFFFFFFFF
    result = (op1 * op2) & mask

    cpu.regs.set(rd, result)
    cpu.regs.cpsr.sign_flag = sign_32(result)
    cpu.regs.cpsr.zero_flag = result == 0

    arm_multiply_idle(cpu, op1, True)


def thumb_high_reg_branch_exchange(cpu: CPU, instr: int):
    """Execute a THUMB.5 instruction (high register operations or branch exchange)"""

    rs = (get_bit(instr, 6) << 3) | get_bits(instr, 3, 5)
    rd = (get_bit(instr, 7) << 3) | get_bits(instr, 0, 2)

    opcode = get_bits(instr, 8, 9)

    if opcode == 0:  # ADD
        cpu.regs.set(rd, add_32(cpu.regs.get(rd), cpu.regs.get(rs)))
        if rd == cpu.regs.PC:
            cpu.flush_pipeline()
            return

    elif opcode == 1:  # CMP
        arm_alu_cmp(cpu, cpu.regs.get(rd), cpu.regs.get(rs))

    elif opcode == 2:  # MOV
        cpu.regs.set(rd, cpu.regs.get(rs))
        if rd == cpu.regs.PC:
            cpu.flush_pipeline()
            return

    elif opcode == 3:  # BX
        address = cpu.regs.get(rs)

        # Mask out the last bit indicating whether to switch to ARM mode
        cpu.regs.pc = address & ~1
        cpu.regs.cpsr.state = get_bit(address, 0)

        cpu.flush_pipeline()
        return

    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def thumb_get_address(cpu: CPU, instr: int):
    """Execute a THUMB.12 instruction (get address relative to SP/PC)"""

    opcode = get_bit(instr, 11)
    rd = get_bits(instr, 8, 10)
    offset = get_bits(instr, 0, 7) * 4

    if opcode:
        # Load from SP
        cpu.regs.set(rd, add_32(cpu.regs.sp, offset))
    else:
        # Load from PC
        cpu.regs.set(rd, add_32(cpu.regs.pc & ~2, offset))

    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def thumb_add_offset_to_stack_pointer(cpu: CPU, instr: int):
    """Execute a THUMB.13 instruction (add offset to SP)"""

    sign = -1 if get_bit(instr, 7) else 1
    offset = sign * get_bits(instr, 0, 6) * 4

    cpu.regs.sp = add_32(cpu.regs.sp, offset)

    cpu.advance_pc_thumb()
    cpu.next_fetch_access = MemoryAccess.SEQUENTIAL



cpdef enum Interrupt:
    VBLANK = 1
    HBLANK = 2
    VCOUNT = 4
    TIMER_0 = 8
    TIMER_1 = 16
    TIMER_2 = 32
    TIMER_3 = 64
    SERIAL = 128
    DMA_0 = 256
    DMA_1 = 512
    DMA_2 = 1024
    DMA_3 = 2048
    KEYPAD = 4096
    GAMEPAK = 8192
    ALL = 16383

cpdef enum PowerDownMode:
    NONE = 0
    HALT = 1
    STOP = 2

cpdef enum EventTrigger:
    IMMEDIATELY = 0
    EventTrigger_VBLANK = 1
    EventTrigger_HBLANK = 2

cpdef enum Key:
    BUTTON_A = 1
    BUTTON_B = 2
    BUTTON_SELECT = 4
    BUTTON_START = 8
    DPAD_RIGHT = 16
    DPAD_LEFT = 32
    DPAD_UP = 64
    DPAD_DOWN = 128
    SHOULDER_RIGHT = 256
    SHOULDER_LEFT = 512

cdef enum:
    CLOCK_SPEED_HZ = 16777216
    NANOSECONDS_PER_SECOND = 1000000000




# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.utils import extend_sign_24, get_bit, add_32
# endif


def arm_branch(cpu: CPU, instr: int):
    """Execute a Branch or Branch and Link instruction"""

    offset = extend_sign_24((instr & 0xFFFFFF)) * 4
    link_flag = get_bit(instr, 24)

    if link_flag:
        cpu.regs.lr = add_32(cpu.regs.pc, -4)

    cpu.regs.pc = add_32(cpu.regs.pc, offset)
    cpu.flush_pipeline()


def arm_branch_exchange(cpu: CPU, instr: int):
    """Execute a Branch and Exchange instruction"""

    address = cpu.regs.get(instr & 0xF)

    # Mask out the last bit indicating whether to switch to THUMB mode
    cpu.regs.pc = address & ~1
    cpu.regs.cpsr.state = get_bit(address, 0)
    cpu.flush_pipeline()



from __future__ import annotations
from enum import IntEnum
DISPLAY_WIDTH = 240
DISPLAY_HEIGHT = 160
HBLANK_PIXELS = 68
VBLANK_LINES = 68
DISPLAY_FULL_WIDTH = DISPLAY_WIDTH + HBLANK_PIXELS
DISPLAY_FULL_HEIGHT = DISPLAY_HEIGHT + VBLANK_LINES
CYCLES_PIXEL = 4
CYCLES_HDRAW = DISPLAY_WIDTH * CYCLES_PIXEL
CYCLES_HBLANK = HBLANK_PIXELS * CYCLES_PIXEL
CYCLES_FRAME = DISPLAY_FULL_WIDTH * DISPLAY_FULL_HEIGHT * CYCLES_PIXEL
SMALL_DISPLAY_WIDTH = 160
SMALL_DISPLAY_HEIGHT = 128
MAX_NUM_OBJECTS = 128
OAM_ENTRY_SIZE = 8
TILE_WIDTH = 8
TILE_HEIGHT = 8
TILE_SIZE = TILE_WIDTH * TILE_HEIGHT
TILE_MAP_ENTRY_SIZE = 2
TILE_MAP_BLOCK_WIDTH = 32
TILE_MAP_BLOCK_HEIGHT = 32
TILE_MAP_BLOCK_SIZE = TILE_MAP_BLOCK_WIDTH * TILE_MAP_BLOCK_HEIGHT * TILE_MAP_ENTRY_SIZE
TILE_SET_BLOCK_SIZE = TILE_MAP_BLOCK_SIZE * 8
COLOUR_SIZE = 2
TRANSPARENT_COLOUR = 32768
OBJ_TILE_SET_OFFSET = 65536
OBJ_PALETTE_OFFSET = 512
VRAM_PAGE_OFFSET = 40960
NUM_BACKGROUNDS = 4
NUM_PRIORITIES = 4
NUM_PRIMARY_WINDOWS = 2

class LayerType(IntEnum):
    BG_0 = 0
    BG_1 = 1
    BG_2 = 2
    BG_3 = 3
    OBJ = 4
    BACKDROP = 5

class WindowIndex(IntEnum):
    WIN_0 = 0
    WIN_1 = 1
    WIN_OBJ = 2
    WIN_OUT = 3

class VideoMode(IntEnum):
    MODE_0 = 0
    MODE_1 = 1
    MODE_2 = 2
    MODE_3 = 3
    MODE_4 = 4
    MODE_5 = 5

class ObjectMode(IntEnum):
    NORMAL = 0
    BLEND = 1
    WINDOW = 2

class ObjectShape(IntEnum):
    SQUARE = 0
    HORIZONTAL = 1
    VERTICAL = 2

class BlendMode(IntEnum):
    OFF = 0
    ALPHA = 1
    LIGHTEN = 2
    DARKEN = 3


class DisplayControlRegister:

    def __init__(self):
        self.reg = 0

    def get_video_mode(self) -> VideoMode | int:
        return get_bits(self.reg, 0, 2)

    def get_page_select(self) -> bint:
        return get_bit(self.reg, 4)

    def get_hblank_interval_free(self) -> bint:
        return get_bit(self.reg, 5)

    def get_obj_vram_dimension(self) -> bint:
        return get_bit(self.reg, 6)

    def get_force_blank(self) -> bint:
        return get_bit(self.reg, 7)

    def display_bg(self, bg_num: int) -> bint:
        return get_bit(self.reg, 8 + bg_num)

    def get_display_obj(self) -> bint:
        return get_bit(self.reg, 12)

    def display_window(self, window: WindowIndex) -> bint:
        return get_bit(self.reg, 13 + window)

class DisplayStatusRegister:

    def __init__(self):
        self.reg = 0

    def get_vblank_status(self) -> bint:
        return get_bit(self.reg, 0)

    def set_vblank_status(self, value: bint):
        self.reg = set_bit(self.reg, 0, value)

    def get_hblank_status(self) -> bint:
        return get_bit(self.reg, 1)

    def set_hblank_status(self, value: bint):
        self.reg = set_bit(self.reg, 1, value)

    def get_vcount_trigger_status(self) -> bint:
        return get_bit(self.reg, 2)

    def set_vcount_trigger_status(self, value: bint):
        self.reg = set_bit(self.reg, 2, value)

    def get_vblank_irq(self) -> bint:
        return get_bit(self.reg, 3)

    def set_vblank_irq(self, value: bint):
        self.reg = set_bit(self.reg, 3, value)

    def get_hblank_irq(self) -> bint:
        return get_bit(self.reg, 4)

    def set_hblank_irq(self, value: bint):
        self.reg = set_bit(self.reg, 4, value)

    def get_vcount_irq(self) -> bint:
        return get_bit(self.reg, 5)

    def set_vcount_irq(self, value: bint):
        self.reg = set_bit(self.reg, 5, value)

    def get_vcount_trigger_value(self) -> int:
        return get_bits(self.reg, 8, 15)

    def set_vcount_trigger_value(self, value: int):
        self.reg = self.reg & 255 | value << 8

class BackgroundControlRegister:

    def __init__(self):
        self.reg = 0

    def get_priority(self) -> int:
        return get_bits(self.reg, 0, 1)

    def get_tile_set_block(self) -> int:
        return get_bits(self.reg, 2, 3)

    def get_mosaic(self) -> bint:
        return get_bit(self.reg, 6)

    def get_colour_256(self) -> bint:
        return get_bit(self.reg, 7)

    def get_tile_map_block(self) -> int:
        return get_bits(self.reg, 8, 12)

    def get_wraparound(self) -> bint:
        return get_bit(self.reg, 13)

    def get_size(self) -> int:
        return get_bits(self.reg, 14, 15)

class WindowControlRegister:

    def __init__(self):
        self.reg = 0

    def display_layer(self, layer_type: LayerType) -> bint:
        return get_bit(self.reg, layer_type)

    def get_enable_blending(self) -> bint:
        return get_bit(self.reg, 5)

class BlendControlRegister:

    def __init__(self):
        self.reg = 0

    def blend_source(self, layer_type: LayerType | int) -> bint:
        return get_bit(self.reg, layer_type)

    def blend_target(self, layer_type: LayerType | int) -> bint:
        return get_bit(self.reg, 8 + layer_type)

    def get_blend_mode(self) -> BlendMode | int:
        return get_bits(self.reg, 6, 7)

class BlendAlphaRegister:

    def __init__(self):
        self.reg = 0

    def get_coefficient_source(self) -> int:
        return get_bits(self.reg, 0, 4)

    def get_coefficient_target(self) -> int:
        return get_bits(self.reg, 8, 12)

class BlendBrightnessRegister:

    def __init__(self):
        self.reg = 0

    def get_brightness(self) -> int:
        return get_bits(self.reg, 0, 4)


from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.constants cimport CPUMode
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport (
    get_bit,
    get_bits,
    set_bit,
    add_32,
)

cdef void arm_block_data_transfer(CPU, uint32_t) noexcept



from libc.stdint cimport uint32_t

from pyboy_advance.constants cimport EventTrigger, Interrupt
from pyboy_advance.interrupt_controller cimport InterruptController
from pyboy_advance.scheduler cimport Scheduler, Event
from pyboy_advance.utils cimport get_bit, get_bits, set_bit

cdef class Timers:
    cdef Timer timer_0
    cdef Timer timer_1
    cdef Timer timer_2
    cdef Timer timer_3

cdef class Timer:
    cdef int WRITE_TIMER_REGISTER_DELAY
    cdef int START_TIMER_DELAY
    cdef int[4] FREQUENCIES

    cdef Scheduler scheduler
    cdef InterruptController interrupt_controller
    cdef int interrupt
    cdef Timer next_timer
    cdef uint32_t _counter
    cdef uint32_t _reload_value
    cdef uint32_t _control_reg
    cdef uint32_t _pending_reload_value
    cdef uint32_t _pending_control_reg
    cdef Event _overflow_event

    cdef uint32_t get_counter(self) noexcept
    cdef void set_counter(self, uint32_t) noexcept
    cdef uint32_t get_control_reg(self) noexcept
    cdef void set_control_reg(self, uint32_t) noexcept
    cdef int get_frequency(self) noexcept
    cdef bint get_count_up(self) noexcept
    cdef void set_count_up(self, bint) noexcept
    cdef bint get_irq_enabled(self) noexcept
    cdef bint get_timer_enabled(self) noexcept
    cdef void _write_reload_value(self) noexcept
    cdef void _write_control_reg(self) noexcept
    cdef void _start(self) noexcept
    cdef void _stop(self) noexcept
    cdef void _update_counter(self) noexcept
    cdef void _schedule_overflow_event(self, uint32_t) noexcept
    cdef void _overflow(self) noexcept



import ast
from dataclasses import dataclass
from os import PathLike
from pathlib import Path


def preprocess_properties(file_paths: list[str | PathLike]) -> list[str | PathLike]:
    """
    Convert methods decorated with @property and @property.setter to explicit getter
    and setter methods, and replace all property usages with method calls.

    Cython does not optimize @property/@property.setter to pure C, which is why
    this transformation is necessary.
    """
    file_asts: dict[Path, ast.Module] = {}
    for file_path in file_paths:
        with open(file_path, "r", encoding="utf-8") as f:
            source_code = f.read()
            file_asts[Path(file_path)] = ast.parse(source_code)

    collector = PropertyCollector()
    for file, tree in file_asts.items():
        if "api" not in file.parts:
            collector.current_class = None
            collector.visit(tree)

    for file, tree in file_asts.items():
        transformer = PropertyTransformer(collector.class_properties)
        tree = transformer.visit(tree)

        replacer = PropertyUsageReplacer(collector.class_properties)
        tree = replacer.visit(tree)

        ast.fix_missing_locations(tree)

        preprocessed_code = ast.unparse(tree)

        with open(file, "w", encoding="utf-8") as f:
            f.write(preprocessed_code)

    return file_paths


@dataclass
class PropertyInfo:
    name: str
    getter_lineno: int
    getter_func: ast.FunctionDef | None = None
    setter_func: ast.FunctionDef | None = None


class PropertyCollector(ast.NodeVisitor):
    """Collects all @property decorators and their associated methods."""

    def __init__(self):
        self.current_class = None
        self.class_properties: dict[str, dict[str, PropertyInfo]] = {}

    def visit_ClassDef(self, node: ast.ClassDef):
        previous_class = self.current_class
        self.current_class = node.name

        if self.current_class not in self.class_properties:
            self.class_properties[self.current_class] = {}

        self.generic_visit(node)
        self.current_class = previous_class

    def visit_FunctionDef(self, node: ast.FunctionDef):
        if self.current_class is None:
            return

        for decorator in node.decorator_list:
            # Check for @property
            if isinstance(decorator, ast.Name) and decorator.id == "property":
                prop_info = PropertyInfo(node.name, node.lineno)
                prop_info.getter_func = node
                self.class_properties[self.current_class][node.name] = prop_info

            # Check for @<name>.setter
            elif isinstance(decorator, ast.Attribute):
                if decorator.attr == "setter" and isinstance(decorator.value, ast.Name):
                    prop_name = decorator.value.id
                    self.class_properties[self.current_class][prop_name].setter_func = node


class PropertyUsageReplacer(ast.NodeTransformer):
    """Replaces property accesses with getter/setter method calls."""

    def __init__(self, properties: dict[str, dict[str, PropertyInfo]]):
        self.current_class = None
        self.property_names = set(
            [prop for class_props in properties.values() for prop in class_props.keys()]
        )

    def visit_ClassDef(self, node: ast.ClassDef):
        previous_class = self.current_class
        self.current_class = node.name
        self.generic_visit(node)
        self.current_class = previous_class
        return node

    def visit_Assign(self, node: ast.Assign):
        # Check for property assignment: obj.property_name = value
        if (
            len(node.targets) == 1
            and isinstance(node.targets[0], ast.Attribute)
            and node.targets[0].attr in self.property_names
        ):
            attr = node.targets[0]
            setter_name = f"set_{attr.attr}"

            # Visit the value side to replace any property accesses there
            new_value = self.visit(node.value)

            # Replace with setter call: obj.set_property_name(value)
            return ast.Expr(
                value=ast.Call(
                    func=ast.Attribute(
                        value=attr.value,
                        attr=setter_name,
                        ctx=ast.Load(),
                    ),
                    args=[new_value],
                    keywords=[],
                )
            )

        self.generic_visit(node)
        return node

    def visit_AugAssign(self, node: ast.AugAssign):
        # Check for property augmented assignment (e.g. obj.property_name += value)
        if isinstance(node.target, ast.Attribute) and node.target.attr in self.property_names:
            attr = node.target
            prop_name = attr.attr

            getter_name = f"get_{prop_name}"
            setter_name = f"set_{prop_name}"

            # obj.get_prop()
            getter_call = ast.Call(
                func=ast.Attribute(
                    value=attr.value,
                    attr=getter_name,
                    ctx=ast.Load(),
                ),
                args=[],
                keywords=[],
            )

            # Visit the value side to replace any property accesses there
            value = self.visit(node.value)

            # Build: obj.get_prop() <op> value
            new_value = ast.BinOp(
                left=getter_call,
                op=node.op,
                right=value,
            )

            # Replace with: obj.set_prop(obj.get_prop() <op> value)
            return ast.Expr(
                value=ast.Call(
                    func=ast.Attribute(
                        value=attr.value,
                        attr=setter_name,
                        ctx=ast.Load(),
                    ),
                    args=[new_value],
                    keywords=[],
                )
            )

        return self.generic_visit(node)

    def visit_Attribute(self, node: ast.Attribute):
        self.generic_visit(node)

        # Check if this is a property access (e.g. self.property_name or obj.property_name)
        if isinstance(node.ctx, ast.Load) and node.attr in self.property_names:
            # Replace with getter call: obj.get_property_name()
            getter_name = f"get_{node.attr}"
            return ast.Call(
                func=ast.Attribute(
                    value=node.value,
                    attr=getter_name,
                    ctx=ast.Load(),
                ),
                args=[],
                keywords=[],
            )

        return node


class PropertyTransformer(ast.NodeTransformer):
    """Transforms @property decorators into explicit getter/setter methods."""

    def __init__(self, properties: dict[str, dict[str, PropertyInfo]]):
        self.properties = properties
        self.current_class = None

    def visit_ClassDef(self, node: ast.ClassDef):
        previous_class = self.current_class
        self.current_class = node.name

        if node.name in self.properties:
            new_body = []

            for item in node.body:
                if isinstance(item, ast.FunctionDef) and item.name in self.properties[node.name]:
                    prop_info = self.properties[node.name][item.name]
                    if prop_info.getter_func and prop_info.getter_func == item:
                        getter_method = self._create_getter_method(item)
                        new_body.append(getter_method)
                    elif prop_info.setter_func and prop_info.setter_func == item:
                        setter_method = self._create_setter_method(prop_info.setter_func, item.name)
                        new_body.append(setter_method)
                    else:
                        new_body.append(item)
                else:
                    new_body.append(item)

            node.body = new_body

        self.generic_visit(node)
        self.current_class = previous_class
        return node

    def _create_getter_method(self, original_func: ast.FunctionDef):
        getter_name = f"get_{original_func.name}"

        # Remove @property decorator
        decorators = [d for d in original_func.decorator_list if not self._is_property_decorator(d)]

        return ast.FunctionDef(
            name=getter_name,
            args=original_func.args,
            body=original_func.body,
            decorator_list=decorators,
            returns=original_func.returns,
            type_comment=original_func.type_comment,
        )

    def _create_setter_method(self, setter_func: ast.FunctionDef, property_name: str):
        setter_name = f"set_{property_name}"

        # Remove @property.setter decorator
        decorators = [
            d for d in setter_func.decorator_list if not self._is_property_setter_decorator(d)
        ]

        return ast.FunctionDef(
            name=setter_name,
            args=setter_func.args,
            body=setter_func.body,
            decorator_list=decorators,
            returns=setter_func.returns,
            type_comment=setter_func.type_comment,
        )

    @staticmethod
    def _is_property_decorator(decorator: ast.expr) -> bool:
        return isinstance(decorator, ast.Name) and decorator.id == "property"

    @staticmethod
    def _is_property_setter_decorator(decorator: ast.expr) -> bool:
        return (
            isinstance(decorator, ast.Attribute)
            and decorator.attr == "setter"
            and isinstance(decorator.value, ast.Name)
        )



cimport cython

from libc.stdint cimport uint8_t, uint32_t
from cpython.array cimport array

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.constants cimport CPUState
from pyboy_advance.memory.constants cimport MemoryRegion, MemoryAccess
from pyboy_advance.memory.gamepak cimport GamePak
from pyboy_advance.memory.io cimport IO
from pyboy_advance.scheduler cimport Scheduler
from pyboy_advance.utils cimport (
    get_bit,
    array_read_16,
    array_read_32,
    array_write_32,
    array_write_16,
    ror_32,
    extend_sign_16,
    extend_sign_8,
    get_bits,
)

cdef class Memory:
    cdef Scheduler scheduler
    cdef CPU cpu
    cdef IO io
    cdef uint8_t[:] bios
    cdef uint8_t[:] ewram
    cdef uint8_t[:] iwram
    cdef GamePak gamepak
    cdef uint32_t bios_last_opcode
    cdef WaitstateControlRegister wait_control
    cdef int[2][16] access_time_32
    cdef int[2][16] access_time_16

    cdef uint32_t read_32(self, uint32_t, int) noexcept
    cdef uint32_t read_32_ror(self, uint32_t, int) noexcept
    cdef uint32_t read_16(self, uint32_t, int) noexcept
    cdef uint32_t read_16_signed(self, uint32_t, int) noexcept
    cdef uint32_t read_16_ror(self, uint32_t, int) noexcept
    cdef uint32_t read_8(self, uint32_t, int) noexcept
    cdef uint32_t read_8_signed(self, uint32_t, int) noexcept
    cdef void write_32(self, uint32_t, uint32_t, int) noexcept
    cdef void write_16(self, uint32_t, uint32_t, int) noexcept
    cdef void write_8(self, uint32_t, uint32_t, int) noexcept
    cdef uint32_t _read_32_internal(self, uint32_t, int) noexcept
    cdef uint32_t _read_16_internal(self, uint32_t, int) noexcept
    cdef uint32_t _read_8_internal(self, uint32_t, int) noexcept
    cdef void _write_32_internal(self, uint32_t, uint32_t, int) noexcept
    cdef void _write_16_internal(self, uint32_t, uint32_t, int) noexcept
    cdef void _write_8_internal(self, uint32_t, uint32_t, int) noexcept
    cdef uint32_t _read_unused_memory(self) noexcept
    cdef void _init_access_times(self) noexcept
    cdef void update_waitstates(self) noexcept

    @cython.locals(cycles=int)
    cdef void _idle_for_access(self, uint32_t, int, int) noexcept

cdef class WaitstateControlRegister:
    cdef int[4] NON_SEQUENTIAL_CYCLES
    cdef uint32_t reg

    cdef int get_sram(self) noexcept
    cdef int get_ws0_non_seq(self) noexcept
    cdef int get_ws0_seq(self) noexcept
    cdef int get_ws1_non_seq(self) noexcept
    cdef int get_ws1_seq(self) noexcept
    cdef int get_ws2_non_seq(self) noexcept
    cdef int get_ws2_seq(self) noexcept


from enum import IntFlag, IntEnum


CLOCK_SPEED_HZ = 16777216
NANOSECONDS_PER_SECOND = 1000000000


class Interrupt(IntFlag):
    # fmt: off
    VBLANK      = 1 << 0
    HBLANK      = 1 << 1
    VCOUNT      = 1 << 2
    TIMER_0     = 1 << 3
    TIMER_1     = 1 << 4
    TIMER_2     = 1 << 5
    TIMER_3     = 1 << 6
    SERIAL      = 1 << 7
    DMA_0       = 1 << 8
    DMA_1       = 1 << 9
    DMA_2       = 1 << 10
    DMA_3       = 1 << 11
    KEYPAD      = 1 << 12
    GAMEPAK     = 1 << 13
    ALL         = 0x3FFF
    # fmt: on


class PowerDownMode(IntEnum):
    NONE = 0
    HALT = 1
    STOP = 2


class EventTrigger(IntEnum):
    IMMEDIATELY = 0
    VBLANK = 1
    HBLANK = 2


class Key(IntFlag):
    # fmt: off
    BUTTON_A        = 1 << 0
    BUTTON_B        = 1 << 1
    BUTTON_SELECT   = 1 << 2
    BUTTON_START    = 1 << 3
    DPAD_RIGHT      = 1 << 4
    DPAD_LEFT       = 1 << 5
    DPAD_UP         = 1 << 6
    DPAD_DOWN       = 1 << 7
    SHOULDER_RIGHT  = 1 << 8
    SHOULDER_LEFT   = 1 << 9
    # fmt: on



from pyboy_advance.app.constants cimport WindowEvent
from pyboy_advance.ppu.constants cimport DISPLAY_WIDTH, DISPLAY_HEIGHT, COLOUR_SIZE

cdef class Window:
    cdef list _events
    cdef dict _key_down_map
    cdef dict _key_up_map
    cdef int _window_scale
    cdef object _sdl_window
    cdef object _sdl_renderer
    cdef object _sdl_texture_buffer

    cdef bint get_fullscreen(self) noexcept
    cdef void set_fullscreen(self, bint) noexcept
    cdef list get_events(self) noexcept
    cdef void render(self, c_void_p) noexcept



from __future__ import annotations

from enum import IntEnum


class MemoryAccess(IntEnum):
    NON_SEQUENTIAL = 0
    SEQUENTIAL = 1


class MemoryRegion(IntEnum):
    BIOS_START = 0x00000000
    BIOS_END = 0x00003FFF
    BIOS_REGION = BIOS_START >> 24
    BIOS_MASK = BIOS_END - BIOS_START
    BIOS_SIZE = BIOS_END - BIOS_START + 1

    EWRAM_START = 0x02000000
    EWRAM_END = 0x0203FFFF
    EWRAM_REGION = EWRAM_START >> 24
    EWRAM_MASK = EWRAM_END - EWRAM_START
    EWRAM_SIZE = EWRAM_END - EWRAM_START + 1

    IWRAM_START = 0x03000000
    IWRAM_END = 0x03007FFF
    IWRAM_REGION = IWRAM_START >> 24
    IWRAM_MASK = IWRAM_END - IWRAM_START
    IWRAM_SIZE = IWRAM_END - IWRAM_START + 1

    IO_START = 0x04000000
    IO_END = 0x040003FF
    IO_REGION = IO_START >> 24
    IO_MASK = IO_END - IO_START
    IO_SIZE = IO_END - IO_START + 1

    PALRAM_START = 0x05000000
    PALRAM_END = 0x050003FF
    PALRAM_REGION = PALRAM_START >> 24
    PALRAM_MASK = PALRAM_END - PALRAM_START
    PALRAM_SIZE = PALRAM_END - PALRAM_START + 1

    VRAM_START = 0x06000000
    VRAM_END = 0x06017FFF
    VRAM_REGION = VRAM_START >> 24
    VRAM_MASK_1 = 0x00017FFF
    VRAM_MASK_2 = 0x0001FFFF
    VRAM_SIZE = VRAM_END - VRAM_START + 1

    OAM_START = 0x07000000
    OAM_END = 0x070003FF
    OAM_REGION = OAM_START >> 24
    OAM_MASK = OAM_END - OAM_START
    OAM_SIZE = OAM_END - OAM_START + 1

    GAMEPAK_0_START = 0x08000000
    GAMEPAK_0_END = 0x09FFFFFF
    GAMEPAK_0_REGION_1 = GAMEPAK_0_START >> 24
    GAMEPAK_0_REGION_2 = GAMEPAK_0_END >> 24

    GAMEPAK_1_START = 0x0A000000
    GAMEPAK_1_END = 0x0BFFFFFF
    GAMEPAK_1_REGION_1 = GAMEPAK_1_START >> 24
    GAMEPAK_1_REGION_2 = GAMEPAK_1_END >> 24

    GAMEPAK_2_START = 0x0C000000
    GAMEPAK_2_END = 0x0DFFFFFF
    GAMEPAK_2_REGION_1 = GAMEPAK_2_START >> 24
    GAMEPAK_2_REGION_2 = GAMEPAK_2_END >> 24

    GAMEPAK_MASK = GAMEPAK_0_END - GAMEPAK_0_START
    GAMEPAK_SIZE = GAMEPAK_0_END - GAMEPAK_0_START + 1
    GAMEPAK_REGION_START = GAMEPAK_0_START >> 24
    GAMEPAK_REGION_END = GAMEPAK_2_END >> 24

    SRAM_START = 0x0E000000
    SRAM_END = 0x0E007FFF
    SRAM_REGION = SRAM_START >> 24
    SRAM_MASK = SRAM_END - SRAM_START
    SRAM_SIZE = SRAM_END - SRAM_START + 1


class IOAddress(IntEnum):
    # Display Control Registers
    REG_DISPCNT = 0x04000000
    REG_DISPSTAT = 0x04000004
    REG_VCOUNT = 0x04000006

    # Background Control Registers
    REG_BG0CNT = 0x04000008
    REG_BG1CNT = 0x0400000A
    REG_BG2CNT = 0x0400000C
    REG_BG3CNT = 0x0400000E

    # Background Offset Registers
    REG_BG0HOFS = 0x04000010
    REG_BG0VOFS = 0x04000012
    REG_BG1HOFS = 0x04000014
    REG_BG1VOFS = 0x04000016
    REG_BG2HOFS = 0x04000018
    REG_BG2VOFS = 0x0400001A
    REG_BG3HOFS = 0x0400001C
    REG_BG3VOFS = 0x0400001E

    # Window Registers
    REG_WIN0H = 0x04000040
    REG_WIN1H = 0x04000042
    REG_WIN0V = 0x04000044
    REG_WIN1V = 0x04000046
    REG_WININ = 0x04000048
    REG_WINOUT = 0x0400004A

    # Special Effects Registers
    REG_BLDCNT = 0x04000050
    REG_BLDALPHA = 0x04000052
    REG_BLDY = 0x04000054

    # Sound Control Registers
    REG_SOUNDBIAS = 0x04000088
    REG_FIFO_A = 0x040000A0
    REG_FIFO_B = 0x040000A4

    # DMA Source/Destination Registers
    REG_DMA0SAD_L = 0x040000B0
    REG_DMA0SAD_H = 0x040000B2
    REG_DMA0DAD_L = 0x040000B4
    REG_DMA0DAD_H = 0x040000B6
    REG_DMA1SAD_L = 0x040000BC
    REG_DMA1SAD_H = 0x040000BE
    REG_DMA1DAD_L = 0x040000C0
    REG_DMA1DAD_H = 0x040000C2
    REG_DMA2SAD_L = 0x040000C8
    REG_DMA2SAD_H = 0x040000CA
    REG_DMA2DAD_L = 0x040000CC
    REG_DMA2DAD_H = 0x040000CE
    REG_DMA3SAD_L = 0x040000D4
    REG_DMA3SAD_H = 0x040000D6
    REG_DMA3DAD_L = 0x040000D8
    REG_DMA3DAD_H = 0x040000DA

    # DMA Control Registers
    REG_DMA0CNT_L = 0x040000B8
    REG_DMA0CNT_H = 0x040000BA
    REG_DMA1CNT_L = 0x040000C4
    REG_DMA1CNT_H = 0x040000C6
    REG_DMA2CNT_L = 0x040000D0
    REG_DMA2CNT_H = 0x040000D2
    REG_DMA3CNT_L = 0x040000DC
    REG_DMA3CNT_H = 0x040000DE

    # Timer Registers
    REG_TM0CNT_L = 0x04000100
    REG_TM0CNT_H = 0x04000102
    REG_TM1CNT_L = 0x04000104
    REG_TM1CNT_H = 0x04000106
    REG_TM2CNT_L = 0x04000108
    REG_TM2CNT_H = 0x0400010A
    REG_TM3CNT_L = 0x0400010C
    REG_TM3CNT_H = 0x0400010E

    # Keypad Registers
    REG_KEYINPUT = 0x04000130
    REG_KEYCNT = 0x04000132

    # Interrupt Registers
    REG_IE = 0x04000200
    REG_IF = 0x04000202
    REG_WAITCNT = 0x04000204
    REG_IME = 0x04000208
    REG_HALTCNT = 0x04000300


class DMAAddressAdjustment(IntEnum):
    INCREMENT = 0
    DECREMENT = 1
    LEAVE_UNCHANGED = 2
    INCREMENT_RELOAD = 3


class DMATransferSize(IntEnum):
    HALFWORD = 0
    WORD = 1


class DMAStartTiming(IntEnum):
    IMMEDIATELY = 0
    VBLANK = 1
    HBLANK = 2
    SPECIAL = 3



THUMB_PATTERNS: list[InstrPattern] = [InstrPattern(255, 223, thumb_software_interrupt), InstrPattern(248, 224, thumb_unconditional_branch), InstrPattern(240, 208, thumb_conditional_branch), InstrPattern(240, 240, thumb_long_branch_with_link), InstrPattern(240, 192, thumb_multiple_load_store), InstrPattern(246, 180, thumb_push_pop_registers), InstrPattern(255, 176, thumb_add_offset_to_stack_pointer), InstrPattern(240, 160, thumb_get_address), InstrPattern(240, 144, thumb_sp_relative_load_store), InstrPattern(240, 128, thumb_load_store_halfword), InstrPattern(224, 96, thumb_load_store_immediate_offset), InstrPattern(242, 80, thumb_load_store_register_offset), InstrPattern(242, 82, thumb_load_store_sign_extended), InstrPattern(248, 72, thumb_pc_relative_load), InstrPattern(252, 68, thumb_high_reg_branch_exchange), InstrPattern(252, 64, thumb_alu), InstrPattern(224, 32, thumb_move_compare_add_subtract), InstrPattern(248, 24, thumb_add_subtract), InstrPattern(224, 0, thumb_move_shifted_register)]

def fill_thumb_lut():
    for opcode in range(2 ** 8):
        for pattern in THUMB_PATTERNS:
            if opcode & pattern.mask == pattern.value:
                THUMB_LUT[opcode] = pattern.handler
                break
fill_thumb_lut()

def thumb_decoder(instr: int) -> InstrHandler:
    instruction_handler = THUMB_LUT[instr >> 8]
    return instruction_handler


import os
from time import sleep

class PyBoyAdvance:

    def __init__(self, rom: GamePak | str | os.PathLike, bios: str | os.PathLike | None=None, skip_bios: bool=False, emulation_speed: float=0):
        self.gamepak = GamePak.from_file(rom) if isinstance(rom, (str, os.PathLike)) else rom
        if bios:
            with open(bios, 'rb') as bios_file:
                bios_data = bytearray(bios_file.read())
        else:
            bios_data = b''
        self.scheduler = Scheduler()
        self.memory = Memory(self.scheduler, self.gamepak, bios_data)
        self.interrupt_controller = InterruptController(self.scheduler)
        self.dma_controller = DMAController(self.scheduler, self.memory)
        self.timers = Timers(self.scheduler, self.interrupt_controller)
        self.ppu = PPU(self.scheduler, self.interrupt_controller)
        self.keypad = Keypad(self.interrupt_controller)
        self.memory.io = IO(self.memory, self.interrupt_controller, self.dma_controller, self.timers, self.ppu, self.keypad)
        self.cpu = CPU(self.scheduler, self.memory)
        self.cpu.arm_decoder = arm_decoder
        self.cpu.thumb_decoder = thumb_decoder
        self._time_per_frame = 0
        self._last_frame_time = 0
        self._accumulated_time = 0
        self.set_emulation_speed(emulation_speed)
        self.scheduler.schedule(self._frame_limiter, CYCLES_FRAME, EventTrigger.IMMEDIATELY)
        self._frame_overshoot = 0
        if skip_bios:
            self.cpu.regs.banked_sp[int(BankIndex.BANK_SYSTEM_USER)] = 50364160
            self.cpu.regs.banked_sp[int(BankIndex.BANK_FIQ)] = 50364160
            self.cpu.regs.banked_sp[int(BankIndex.BANK_IRQ)] = 50364320
            self.cpu.regs.banked_sp[int(BankIndex.BANK_SWI)] = 50364384
            self.cpu.regs.banked_sp[int(BankIndex.BANK_ABORT)] = 50364160
            self.cpu.regs.banked_sp[int(BankIndex.BANK_UNDEFINED)] = 50364160
            self.cpu.regs.set_sp(50364160)
            self.cpu.regs.set_pc(134217728)
            self.cpu.flush_pipeline()
        else:
            self.cpu.interrupt(ExceptionVector.RESET)
        self.screen = Screen(self.ppu)

    def step(self):
        if self.dma_controller.get_active():
            self.dma_controller.perform_transfers()
        else:
            self.cpu.step()

    def frame(self, count: int=1, render: bool=True):
        self.ppu.rendering_enabled = False
        for frame_num in range(count):
            if frame_num == count - 1 and render:
                self.ppu.rendering_enabled = True
            end_time = self.scheduler.cycles + CYCLES_FRAME - self._frame_overshoot
            while self.scheduler.cycles < end_time:
                self.step()
            self._frame_overshoot = self.scheduler.cycles - end_time

    def run(self):
        with Window() as window:
            running = True
            while running:
                self.frame()
                for event in window.get_events():
                    if event == WindowEvent.WindowEvent_NONE:
                        continue
                    elif event == WindowEvent.QUIT:
                        running = False
                    elif event == WindowEvent.FULLSCREEN:
                        window.set_fullscreen(not window.get_fullscreen())
                    else:
                        self.keypad.process_window_event(event)
                window.render(self.ppu.get_frame_buffer_ptr())

    def press_key(self, key: Key):
        self.keypad.press_key(key)

    def release_key(self, key: Key):
        self.keypad.release_key(key)

    def set_emulation_speed(self, speed: float):
        if speed < 0:
            raise ValueError('Emulation speed must be 0 or greater')
        self._time_per_frame = int(1 / (speed * CLOCK_SPEED_HZ / CYCLES_FRAME) * NANOSECONDS_PER_SECOND) if speed > 0 else 0
        self._last_frame_time = perf_counter_ns()
        self._accumulated_time = 0

    def _frame_limiter(self):
        self.scheduler.schedule(self._frame_limiter, CYCLES_FRAME, EventTrigger.IMMEDIATELY)
        if self._time_per_frame == 0:
            return
        now = perf_counter_ns()
        self._accumulated_time += now - self._last_frame_time
        self._last_frame_time = now
        if self._accumulated_time < self._time_per_frame:
            delay = (self._time_per_frame - self._accumulated_time) / float(NANOSECONDS_PER_SECOND)
            sleep(delay)
        self._accumulated_time -= self._time_per_frame


# ifndef CYTHON
from __future__ import annotations

from pyboy_advance.constants import Interrupt, PowerDownMode, EventTrigger
from pyboy_advance.scheduler import Scheduler
from pyboy_advance.utils import bint
# endif


class InterruptController:
    def __init__(self, scheduler: Scheduler):
        self.WRITE_INTERRUPT_REGISTERS_DELAY = 1
        self.UPDATE_IRQ_LINE_DELAY = 2

        self.scheduler = scheduler

        self._interrupt_enable = 0
        self._interrupt_flags = 0
        self._interrupt_master_enable = 0

        self._pending_interrupt_enable = 0
        self._pending_interrupt_flags = 0
        self._pending_interrupt_master_enable = 0

        self.irq_line = False

        self.power_down_mode = PowerDownMode.NONE

    @property
    def interrupt_enable(self) -> int:
        return self._interrupt_enable

    @interrupt_enable.setter
    def interrupt_enable(self, value: int):
        self._pending_interrupt_enable = value & Interrupt.ALL
        self._schedule_write_interrupt_registers()

    @property
    def interrupt_flags(self) -> int:
        return self._interrupt_flags

    @interrupt_flags.setter
    def interrupt_flags(self, value: int):
        self._pending_interrupt_flags &= ~value
        self._schedule_write_interrupt_registers()

    @property
    def interrupt_master_enable(self) -> bint:
        return self._interrupt_master_enable

    @interrupt_master_enable.setter
    def interrupt_master_enable(self, value: bint):
        self._pending_interrupt_master_enable = value & 0b1
        self._schedule_write_interrupt_registers()

    def signal(self, interrupt: Interrupt):
        self._pending_interrupt_flags |= interrupt
        self._schedule_write_interrupt_registers()

    def _schedule_write_interrupt_registers(self):
        self.scheduler.schedule(
            self._write_interrupt_registers,
            self.WRITE_INTERRUPT_REGISTERS_DELAY,
            EventTrigger.IMMEDIATELY,
        )

    def _write_interrupt_registers(self):
        self._interrupt_enable = self._pending_interrupt_enable
        self._interrupt_flags = self._pending_interrupt_flags
        self._interrupt_master_enable = self._pending_interrupt_master_enable

        interrupt_available = self._interrupt_enable & self._interrupt_flags

        if interrupt_available and self.power_down_mode == PowerDownMode.HALT:
            self.power_down_mode = PowerDownMode.NONE

        new_irq_line = interrupt_available and self._interrupt_master_enable
        if new_irq_line != self.irq_line:
            if new_irq_line:
                self.scheduler.schedule(
                    self._set_irq_line,
                    self.UPDATE_IRQ_LINE_DELAY,
                    EventTrigger.IMMEDIATELY,
                )
            else:
                self.scheduler.schedule(
                    self._reset_irq_line,
                    self.UPDATE_IRQ_LINE_DELAY,
                    EventTrigger.IMMEDIATELY,
                )

    def _set_irq_line(self):
        self.irq_line = True

    def _reset_irq_line(self):
        self.irq_line = False
